/*
gcc main.c -o tiff2c -ltiff
*/

#include <stdio.h>
#include <stdlib.h>
#include <tiffio.h>
#include <string.h>

#define TRUE    1
#define FALSE   0

/*--------------------------------------------------------------------*/
void    printUsage()
{
	fprintf(stderr, "\n");
	fprintf(stderr, "Convert a TIFF file to a C array. Output is a header file which\n");
	fprintf(stderr, "can be compiled into OpenGL source.\n\n");
	fprintf(stderr, "Usage: tiff2c [options] filename.tif\n");
	fprintf(stderr, "  options:\n");
	fprintf(stderr, "    -v <name>        Set variable name for array\n");
	fprintf(stderr, "    -o <filename>    Write output to <filename> (defaults to stdout)\n");
//	fprintf(stderr, "    -rgb             Output RGB\n");
//	fprintf(stderr, "    -rgba            Output RGBA (default)\n");
//	fprintf(stderr, "    -a               Output Alpha only\n");
	fprintf(stderr, "    -w               Set all RGB components to 255\n");
	fprintf(stderr, "    -n <val>         Add noise to alpha channel (range 0-255)\n");
}

/*-- Globals ---------------------------------------------------------*/
char    *tifFilename = 0;
char    *outFilename = 0;
char    *varName	= "tex_0";
FILE	*outfile	= 0;

uint32* raster	= 0;
uint32  w		= 0;
uint32	h		= 0;

enum	{ ModeRGB, ModeRGBA, ModeA };
int		writeMode = ModeRGBA;

int		whiteTex = 0;	/* set RGB components to 255 */
int		addNoise = 0;	/* addNoise to alpha channel */

void    commandLine(int argc, char** argv);
int		loadTifFile(const char* filename);
void	addNoiseToAlpha(int amount);

/*--------------------------------------------------------------------*/
/* Macros for extracting components from the                          */
/* packed ABGR form returned by TIFFReadRGBAImage.                    */
#define AGBRGetR(abgr)  ( (abgr) & 0xff)
#define AGBRGetG(abgr)  (((abgr) >> 8) & 0xff)
#define AGBRGetB(abgr)  (((abgr) >> 16) & 0xff)
#define AGBRGetA(abgr)  (((abgr) >> 24) & 0xff)
#define RGBAGetR(rgba)  (((rgba) >> 24)& 0xff)
#define RGBAGetG(rgba)  (((rgba) >> 16) & 0xff)
#define RGBAGetB(rgba)  (((rgba) >> 8) & 0xff)
#define RGBAGetA(rgba)  ( (rgba) & 0xff)
/*--------------------------------------------------------------------*/
uint32  AGBR2RGBA (uint32 agbr)
{
	uint32  r = AGBRGetR(agbr) << 24;
	uint32  g = AGBRGetG(agbr) << 16;
	uint32  b = AGBRGetB(agbr) << 8 ;
	uint32  a = AGBRGetA(agbr);
	return ( a | g | b | r );
}

/*==============================================================================*/
int main(int argc, char** argv)
{
	raster	= 0;
	w		= 0;
	h		= 0;
	outfile = stdout;

	commandLine(argc, argv);

	if(loadTifFile(tifFilename))
	{
		uint32 i,j;
		if(outFilename)
			outfile = fopen(outFilename, "w");
		else	
			outfile = stdout;
		
		if(addNoise)
			addNoiseToAlpha(addNoise);
			
		if(outfile)
		{
			fprintf(outfile, "enum { %sw = %ld, %sh = %ld };\n", varName, w, varName, h);
			switch(writeMode)
			{
				case ModeRGB:
					fprintf(outfile, "enum { %sComp = 3, %sFormat = GL_RGB };\n", varName, varName);
					fprintf(outfile, "unsigned char %s[%ld][%ld] = {\n", varName, h, w*3);
					break;
				case ModeRGBA:
					fprintf(outfile, "enum { %sComp = 4, %sFormat = GL_RGBA };\n", varName, varName);
					fprintf(outfile, "unsigned char %s[%ld][%ld] = {\n", varName, h, w*4);
					break;
				case ModeA:
					fprintf(outfile, "enum { %sComp = 1, %sFormat = GL_ALPHA };\n", varName, varName);
					fprintf(outfile, "unsigned char %s[%ld][%ld] = {\n", varName, h, w);
					break;
			}
			for(i = 0; i < h; i++)
			{
				fprintf(outfile, "{");
				for(j = 0; j < w; j++)
				{
					switch(writeMode)
					{
						case ModeRGB:
							fprintf(outfile, " %3d, %3d, %3d", 
								(int)AGBRGetR(raster[(i*w)+j]),
								(int)AGBRGetG(raster[(i*w)+j]),
								(int)AGBRGetB(raster[(i*w)+j]) );
							break;
						case ModeRGBA:
							if(whiteTex)
							{
								fprintf(outfile, " %3d, %3d, %3d, %3d", 255, 255, 255,
									(int)AGBRGetA(raster[(i*w)+j]) );
							}
							else
							{
								fprintf(outfile, " %3d, %3d, %3d, %3d", 
									(int)AGBRGetR(raster[(i*w)+j]),
									(int)AGBRGetG(raster[(i*w)+j]),
									(int)AGBRGetB(raster[(i*w)+j]),
									(int)AGBRGetA(raster[(i*w)+j]) );
							}
							break;
						case ModeA:
							fprintf(outfile, " %3d", 
								(int)AGBRGetA(raster[(i*w)+j]) );
							break;
					}
					if(j != w-1)
						fprintf(outfile, ",");
				}
				if(i != h-1)
					fprintf(outfile, "},\n");
				else
					fprintf(outfile, "}\n");
			}
			fprintf(outfile, "};\n");
		}
		else
			fprintf(stderr, "ERROR: could not open %s for write.\n", outFilename);
			
		_TIFFfree(raster);
		fclose(outfile);
	}
	else
		fprintf(stderr, "ERROR: could not open %s for read.\n", tifFilename);

	exit(0);
}

/*--------------------------------------------------------------------*/
int    loadTifFile(const char* filename)
{
	int		retVal	= FALSE;
	TIFF	*tif	= 0;
	size_t	npixels	= 0;

	tif = TIFFOpen(filename, "r");
	if(tif)
	{
		TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
		TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
		npixels = w * h;
		raster = (uint32*) _TIFFmalloc(npixels * sizeof (uint32));
		if (raster != NULL)
			if (TIFFReadRGBAImage(tif, w, h, raster, 0))
				retVal = TRUE;
		TIFFClose(tif);
	}
	return retVal;
}

/*--------------------------------------------------------------------*/
void	addNoiseToAlpha(int amount)
{
	uint32	i,j;
	double	random;
	uint32	diff;
	uint32	oldR;
	uint32	oldG;
	uint32	oldB;
	uint32	oldA;
	uint32	newA;
	
	if(amount > 255 || amount < 0)
		amount = 0;
	
	for(i = 0; i < h; i++)
	{
		for(j = 0; j < w; j++)
		{
			random = (double)rand()/(double)RAND_MAX;
			oldR = AGBRGetR(raster[(i*w)+j]);
			oldG = AGBRGetG(raster[(i*w)+j]);
			oldB = AGBRGetB(raster[(i*w)+j]);
			oldA = AGBRGetA(raster[(i*w)+j]);
			diff = random*amount;
			if(diff < oldA)
				newA = oldA - diff;
			else
				newA = 0;
			raster[(i*w)+j] = (oldR | oldG << 8 | oldB << 16 | newA << 24);
		}
	}
}

/*--------------------------------------------------------------------*/
void    commandLine(int argc, char** argv)
{
	int 	i;
	int		nameLen;
	
	if(argc < 2)
	{
		printUsage();
		exit(0);
	}

	for(i = 1; i < argc; i++)
	{
		if( strchr(argv[i], '-') == argv[i] )
		{
			if(strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0 )
			{
				printUsage();
				exit(0);
			}
//			else if( strcmp(argv[i], "-rgb") == 0 )
//				writeMode = ModeRGB;
//			else if( strcmp(argv[i], "-rgba") == 0 )
//				writeMode = ModeRGBA;
//			else if( strcmp(argv[i], "-a") == 0 )
//				writeMode = ModeA;
			else if( strcmp(argv[i], "-w") == 0 )
				whiteTex = 1;
			else if( strcmp(argv[i], "-n") == 0 && (i+1 < argc) )
			{
				sscanf(argv[i+1], "%d", &addNoise);
				if(addNoise > 255 || addNoise < 0)
					addNoise = 0;
				i++;
			}
			else if( strcmp(argv[i], "-v") == 0 && (i+1 < argc) )
			{
				nameLen = strlen(argv[i+1])+1;
				varName = (char*)malloc( nameLen );
				strcpy(varName, argv[i+1]);
				i++;
			}
			else if( strcmp(argv[i], "-o") == 0 && (i+1 < argc) )
			{
				nameLen = strlen(argv[i+1])+1;
				outFilename = (char*)malloc( nameLen );
				strcpy(outFilename, argv[i+1]);
				i++;
			}
		}
		else
		{
			nameLen = strlen(argv[i])+1;
			tifFilename = (char*)malloc( nameLen );
			strcpy(tifFilename, argv[i]);

			if( !strstr(tifFilename, ".tif") )
			{
				printUsage();
				exit(0);
			}
		}
	}
	
	if(!tifFilename)
	{
		printUsage();
		exit(0);
	}
}








