/* MBATexMap.sl
 *
 * DESCRIPTION:
 *    Tileable texture map shader with raytraced reflections
 *
 * PARAMETERS:
 *    Ka, Kd, Ks, Kr, roughness, specularcolor - the usual meaning.
 *    textureBlend - amount of texture modulation
 *    sScale, tScale - tiled repeat of the texture
 *    texturename - the name of the texture file.
 *    blur, samples - blurred reflection. This is very expensive
 *
 * ANTIALIASING: yes
 *
 * AUTHOR: Mark B. Allan
 *
 * HISTORY:
 *  The ray-tracing bits were lifted from one of Larry Gritz's shaders.
 *  texture filtering from the Vidi 'texbomb' shader
 *
 * The RenderMan (R) Interface Procedures and RIB Protocol are:
 * Copyright 1988, 1989, Pixar.  All rights reserved.
 * RenderMan (R) is a registered trademark of Pixar.
 */


surface MBATexMap ( 
			float	Ka				= 1.00;
			float	Kd				= 0.50;
			float	Ks				= 0.50;
			float	Kr				= 0.00;
			float	roughness		= 0.75;
			color	specularcolor	= 1.00;
			float	textureBlend	= 1.00;
			float	sScale			= 1.00;
			float	tScale			= 1.00;
			string	texturename		= "";
			float	blur 			= 0;
			float	samples 		= 1)
{
	normal	Nf = faceforward (normalize(N),I);	/* Forward facing normal vector */
	vector	IN = normalize(I);					/* normalized incident vector */
    vector	Rdir;								/* Smooth reflection direction */
    vector	uoffset, voffset;					/* Offsets for blur */
    color	ev;									/* Color of the reflections */
    vector	R;									/* Direction to cast the ray */
	color	Ct;									/* Color from texture sample */
	float	ss, tt;
	float	swidth, twidth;
	float	ds, dt;
	float	dsu, dtu;
	float	dsv, dtv;
    uniform float i, j;
	
	/* texture sampling and filtering */
    ss = s * sScale;
    tt = t * tScale;
	swidth = abs(Du(ss) * du) + abs(Dv(ss) * dv);
	twidth = abs(Du(tt) * du) + abs(Dv(tt) * dv);
	dsu = abs(Du(ss)*du);
	if((2*dsu) > swidth) dsu = swidth - dsu;
	dsv = abs(Dv(ss)*dv);
	if((2*dsv) > swidth) dsv = swidth - dsv;
	ds = dsu + dsv;
	dtu = abs(Du(tt)*du);
	if((2*dtu) > twidth) dtu = twidth - dtu;
	dtv = abs(Dv(tt)*dv);
	if((2*dtv) > twidth) dtv = twidth - dtv;
	dt = dtu + dtv;

	if (texturename != "")
		Ct = color texture (texturename, ss, tt, ss+ds, tt, ss, tt+dt, ss+ds, tt+dt);
	else Ct = 1;

    /* Calculate the reflection color */
    if (Kr > 0.001) 
	{
		/* Rdir gets the perfect reflection direction */
		Rdir = normalize (reflect (IN, Nf));
		if (blur > 0) 
		{
	    	/* Construct orthogonal components to Rdir */
	    	uoffset = blur * normalize (point (zcomp(Rdir) - ycomp(Rdir),
					       				xcomp(Rdir) - zcomp(Rdir),
					       				ycomp(Rdir) - xcomp(Rdir)));
	    	voffset = Rdir ^ uoffset;
	    	for (i = 0;  i < samples;  i += 1) 
			{
				for (j = 0;  j < samples;  j += 1) 
				{
		    		/* Add a random offset to the smooth reflection vector */
		    		R = Rdir +
						((i + float random())/samples - 0.5) * uoffset +
						((j + float random())/samples - 0.5) * voffset;
		    		ev += trace (P, normalize(R));
				}
	    	}
	    	ev *= Kr / (samples*samples);
		}
		else
		{
	    	/* No blur, just do a simple trace */
	    	ev = Kr * trace (P, Rdir);
		}
    }

	Oi = Os;
	Ci = Os * ( (((1.0-textureBlend)*Cs) + (Cs*textureBlend*Ct)) * (Ka*ambient() + Kd*diffuse(Nf)) 
				+ specularcolor * (ev + Ks*specular(Nf,-IN,roughness)) );
}

