/**
 * camera.cpp
 * Copyright 1999 by Mark B. Allan
 * 
 * camera class for GL and RM projection
 */
#include "camera.h"

#include "compatibility.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <math.h>

#include "mmath.h"

#include <ri.h>

//==============================================================================
// CAMERAVIEW
//==============================================================================
CameraView::CameraView(int pType,  int gridNumUnits, float gridUnitSize)
{
	viewName	= "";
//	targetLock	= false;

	vals = new CameraVals();
	vals->projType	= pType;
	vals->position.set(0.0, 0.0, 0.0);
	vals->target.set(0.0, 0.0, 0.0);
	vals->rotate.set(0.0, 0.0, 0.0);
	vals->fov 		= 40.0;
	vals->aspect	= 1.0;
	vals->gridX		= false;
	vals->gridY		= true;
	vals->gridZ		= false;
	vals->unitSize	= gridUnitSize;
	vals->numUnits	= gridNumUnits;
	vals->intensity	= 0.2;

	nearPlane	= 1.0;
	farPlane	= 500.0;
	
	hasFog		= false;
	fogType		= Fog;
	fogDensity	= 0.1;
	fogDistance	= 100.0;
	fogNear		= 1.0;
	fogFar		= 10.0;
	fogColor.set(0.0, 0.0, 0.0);
	
	init = new CameraVals(*vals);
}

//----------------------------------------------------------------
CameraView::CameraView (CameraView &in)
	: QObject()
{
	viewName	= in.name();
	viewName	+= "_copy";
//	targetLock	= in.TargetLock();
	vals 		= new CameraVals(in.GetVals());
	init 		= new CameraVals(*vals);
	
	nearPlane	= in.nearPlane;
	farPlane	= in.farPlane;
	
	hasFog		= in.hasFog;
	fogType		= in.fogType;
	fogDensity	= in.fogDensity;
	fogDistance	= in.fogDistance;
	fogNear		= in.fogNear;
	fogFar		= in.fogFar;
	fogColor	= in.fogColor;
}

//----------------------------------------------------------------
CameraView::~CameraView()
{
	delete	init;
	delete	vals;
}

//----------------------------------------------------------------
void	CameraView::SetProjection (int projectionType)
{
	vals->projType = projectionType;
}

//----------------------------------------------------------------
void	CameraView::SetName (const char *name)
{
	viewName = name;
}

//----------------------------------------------------------------
void	CameraView::Reset ()
{
	*vals = *init;
}

//----------------------------------------------------------------
void	CameraView::drawGL()
{
//	glTranslatef(vals->position.X(), vals->position.Y(), vals->position.Z());
	glPushMatrix();
	MMatrix	rot;
	rot.rotateX(-vals->rotate.x() );
	rot.rotateZ(-vals->rotate.z() );
	rot.rotateY(-vals->rotate.y() );
	rot.setTranslate( vals->position.x(), vals->position.y(), vals->position.z());
	glMultMatrixf( rot.mat() );
	float	zoomLen = (1.0-vals->fov/180.0)-0.5;
	glBegin(GL_QUADS);
		//front
		glNormal3f( 0.00, 0.00,-1.00 );
		glVertex3f(	0.25,-0.25,-0.25 );
		glVertex3f(-0.25,-0.25,-0.25 );
		glVertex3f(-0.25, 0.25,-0.25 );
		glVertex3f(	0.25, 0.25,-0.25 );
		//back
		glNormal3f( 0.00, 0.00, 1.00 );
		glVertex3f(	0.25,-0.25, 0.50 );
		glVertex3f(	0.25, 0.25, 0.50 );
		glVertex3f(-0.25, 0.25, 0.50 );
		glVertex3f(-0.25,-0.25, 0.50 );
		//side
		glNormal3f(-1.00, 0.00, 0.00 );
		glVertex3f(-0.25,-0.25,-0.25 );
		glVertex3f(-0.25,-0.25, 0.50 );
		glVertex3f(-0.25, 0.25, 0.50 );
		glVertex3f(-0.25, 0.25,-0.25 );		
		//side
		glNormal3f( 1.00, 0.00, 0.00 );
		glVertex3f( 0.25,-0.25,-0.25 );
		glVertex3f( 0.25, 0.25,-0.25 );
		glVertex3f( 0.25, 0.25, 0.50 );
		glVertex3f( 0.25,-0.25, 0.50 );
		//top
		glNormal3f( 0.00, 1.00, 0.00 );
		glVertex3f(-0.25, 0.25,-0.25 );
		glVertex3f(-0.25, 0.25, 0.50 );
		glVertex3f( 0.25, 0.25, 0.50 );
		glVertex3f( 0.25, 0.25,-0.25 );
		//bottom
		glNormal3f( 0.00,-1.00, 0.00 );
		glVertex3f(-0.25,-0.25,-0.25 );
		glVertex3f( 0.25,-0.25,-0.25 );
		glVertex3f( 0.25,-0.25, 0.50 );
		glVertex3f(-0.25,-0.25, 0.50 );
		//lens
		glNormal3f( 0.00, 0.00,-1.00 );
		glVertex3f(	0.25,-0.25,-0.51-zoomLen );
		glVertex3f(-0.25,-0.25,-0.51-zoomLen );
		glVertex3f(-0.25, 0.25,-0.51-zoomLen );
		glVertex3f(	0.25, 0.25,-0.51-zoomLen );
	glEnd();
	glBegin(GL_TRIANGLES);
		glNormal3f( 1.00, 0.00, 0.00 );
		glVertex3f( 0.00, 0.00,-0.25 );
		glVertex3f( 0.25,-0.25,-0.51-zoomLen );
		glVertex3f( 0.25, 0.25,-0.51-zoomLen );
		
		glNormal3f( 0.00,-1.00, 0.00 );
		glVertex3f( 0.00, 0.00,-0.25 );
		glVertex3f(-0.25,-0.25,-0.51-zoomLen );
		glVertex3f( 0.25,-0.25,-0.51-zoomLen );
		
		glNormal3f(-1.00, 0.00, 0.00 );
		glVertex3f( 0.00, 0.00,-0.25 );
		glVertex3f(-0.25, 0.25,-0.51-zoomLen );
		glVertex3f(-0.25,-0.25,-0.51-zoomLen );
		
		glNormal3f( 0.00, 1.00, 0.00 );
		glVertex3f( 0.00, 0.00,-0.25 );
		glVertex3f( 0.25, 0.25,-0.51-zoomLen );
		glVertex3f(-0.25, 0.25,-0.51-zoomLen );
	glEnd();
	glBegin(GL_LINES);
		glVertex3f( 0.00,-0.25, 0.00 );
		glVertex3f( 0.00,-1.25, 0.00 );
	glEnd();
//	glBegin(GL_LINES);
//		glVertex3f( 0.00, 0.00,-0.25 );
//		glVertex3f( 0.25, 0.25,-0.75 );
//		glVertex3f( 0.00, 0.00,-0.25 );
//		glVertex3f( 0.25,-0.25,-0.75 );
//		glVertex3f( 0.00, 0.00,-0.25 );
//		glVertex3f(-0.25,-0.25,-0.75 );
//		glVertex3f( 0.00, 0.00,-0.25 );
//		glVertex3f(-0.25, 0.25,-0.75 );
//		
//		glVertex3f( 0.25, 0.25,-0.75 );
//		glVertex3f( 0.25,-0.25,-0.75 );
//		
//		glVertex3f( 0.25,-0.25,-0.75 );
//		glVertex3f(-0.25,-0.25,-0.75 );
//		
//		glVertex3f(-0.25,-0.25,-0.75 );
//		glVertex3f(-0.25, 0.25,-0.75 );
//		
//		glVertex3f(-0.25, 0.25,-0.75 );
//		glVertex3f(	0.25, 0.25,-0.75 );
//		
//		glVertex3f( 0.00,-0.25, 0.00 );
//		glVertex3f( 0.00,-1.25, 0.00 );
//	glEnd();
	glPopMatrix();
}

//----------------------------------------------------------------
void	CameraView::ProjectGL()
{
	float	fov		= vals->fov;
	float	aspect	= vals->aspect;
		
	if(vals->projType == CAM_PERSPECTIVE)
	{
    	glMatrixMode( GL_PROJECTION );
    	glLoadIdentity();
		gluPerspective(vals->fov, vals->aspect, nearPlane, farPlane);
		glMatrixMode(GL_MODELVIEW);
	}
	else
	{
    	glMatrixMode( GL_PROJECTION );
    	glLoadIdentity();
		glOrtho(-fov*aspect, vals->fov*aspect, -fov, fov, nearPlane, farPlane);
		glMatrixMode(GL_MODELVIEW);
	}
}

//----------------------------------------------------------------
void	CameraView::PlaceCamGL()
{
	if (vals->targetLock)
		LookatTarget();
	
//	//-- Fog stuff...this has to be moved
//	switch( fogType )
//	{
//		case NoFog:
//			glDisable(GL_FOG);
//			break;
//		case Fog:
//			glEnable(GL_FOG);
//			glFogi ( GL_FOG_MODE, GL_EXP2 );
//			glFogfv( GL_FOG_COLOR, fogColor.XYZ() );
//			glFogf ( GL_FOG_DENSITY, fogDensity );
//			glFogf ( GL_FOG_START, fogNear );
//			glFogf ( GL_FOG_END, fogFar );
//			break;
//		case DepthCue:
//			glEnable(GL_FOG);
//			glFogi ( GL_FOG_MODE, GL_LINEAR );
//			glFogfv( GL_FOG_COLOR, fogColor.XYZ() );
//			glFogf ( GL_FOG_START, fogNear );
//			glFogf ( GL_FOG_END, fogFar );
//			break;
//	}
//	//-- end Fog stuff
	
	MMatrix	rot;
	rot.setTranslate( -vals->position.x(), -vals->position.y(), -vals->position.z());
	rot.rotateY( vals->rotate.y() );
	rot.rotateX( vals->rotate.x() );
	rot.rotateZ( vals->rotate.z() );
	glMultMatrixf( rot.mat() );
	
//	glRotatef ( vals->rotate.Z(), 0.0, 0.0, 1.0);
//	glRotatef ( vals->rotate.X(), 1.0, 0.0, 0.0);
//	glRotatef ( vals->rotate.Y(), 0.0, 1.0, 0.0);
//	glTranslatef(-vals->position.X(), -vals->position.Y(), -vals->position.Z());
}					

//----------------------------------------------------------------
void	CameraView::ProjectRIB()
{
	RtFloat	fovRIB[1] = { (RtFloat)vals->fov	};
	
//	//-- Fog stuff...this has to be moved
//	switch( fogType )
//	{
//		case NoFog:
//			break;
//		case Fog:
//			RiAtmosphere("fog", 
//						"distance", &fogDistance,
//						"background", fogColor.XYZ(), 
//						RI_NULL); 
//			break;
//		case DepthCue:
//			RiAtmosphere("depthcue", 
//						"mindistance", &fogNear, 
//						"maxdistance", &fogFar, 
//						"background", fogColor.XYZ(), 
//						RI_NULL); 
//			break;
//	}
//	//-- end Fog stuff
	
	if(vals->projType == CAM_PERSPECTIVE)
	{
		RiProjection ("perspective", "fov", fovRIB, RI_NULL);
	}
	else if(vals->projType == CAM_ORTHO)
	{
		RiProjection ("orthographic", RI_NULL);
		RiFrameAspectRatio ( (RtFloat) 320.0f/240.0f );
	}
}

//----------------------------------------------------------------
void	CameraView::PlaceCamRIB()
{
	if (vals->targetLock)
		LookatTarget();
	
	MMatrix	rot;
	rot.setTranslate( -vals->position.x(), -vals->position.y(), vals->position.z());
	rot.rotateY( -vals->rotate.y() );
	rot.rotateX( -vals->rotate.x() );
	rot.rotateZ( -vals->rotate.z() );
	rot.RIBConcatTransform();
}
//----------------------------------------------------------------
void	CameraView::LookatTarget()
{
	vals->direction.set(vals->target.x()-vals->position.x(),
						vals->target.y()-vals->position.y(),
						vals->target.z()-vals->position.z() );
	vals->directionXZ = vals->direction;
	vals->directionXZ.setY(0.0);
	vals->directionXZ.normalize();

	LookInDirection();
}

//----------------------------------------------------------------
void	CameraView::LookInDirection()
{	//most of this was lifted from _The_Renderman_Companion_ listing 8.2
	double	xzlen, yzlen;
	double	yrot, xrot;
	
	float	direction[3] = {-vals->direction.x(), -vals->direction.y(), -vals->direction.z()};				
				
	if (direction[0] == 0.0 && direction[1] == 0.0 && direction[2]==0.0 )
		return;
		
	xzlen = sqrt(direction[0]*direction[0]+direction[2]*direction[2]);
	if (xzlen == 0)
		yrot = (direction[1]<0.0) ? 180.0 : 0.0;
	else
		yrot = 180.0*acos(direction[2]/xzlen)/M_PI;
		
	yzlen = sqrt(direction[1]*direction[1]+xzlen*xzlen);
	if (yzlen == 0.0)
		warning("CameraView::SetLookatRotation() - yzlen is 0"), yzlen = 0.0000001;
	xrot = 180.0*acos(xzlen/yzlen)/M_PI;
	
	if (direction[1] > 0.0)
		vals->rotate.setX( xrot);
	else
		vals->rotate.setX(-xrot);
		
	if (direction[0] > 0.0)
		vals->rotate.setY(-yrot);
	else
		vals->rotate.setY( yrot);
}

//----------------------------------------------------------------
void	CameraView::SetFov(float newfov)
{	vals->fov = newfov;	}
float	CameraView::AddFov(float incfov)
{
	vals->fov += incfov;
	if (ProjType() == CAM_PERSPECTIVE)
	{
		if (vals->fov < 1.0)		vals->fov = 1.0;
		if (vals->fov > 179.0)	vals->fov = 179.0;
	}
	else if (ProjType() == CAM_ORTHO)
	{
		if (vals->fov < 0.1)		vals->fov = 0.1;
	}
	return vals->fov;
}

//----------------------------------------------------------------
void	CameraView::SetAspect(float newAspect)
{	vals->aspect = newAspect;	}

//----------------------------------------------------------------
void	CameraView::MoveInOut(float amnt)
{
	float movX;
	float movZ;
	
	movX = -amnt*vals->directionXZ.x();
	movZ = -amnt*vals->directionXZ.z();
	
	vals->position.addX(movX);
	vals->position.addZ(movZ);
}

void	CameraView::MoveRightLeft(float amnt)
{
	float movX;
	float movZ;
	
	movX = -amnt*vals->directionXZ.z();
	movZ = amnt*vals->directionXZ.x();

	vals->position.addX(movX);
	vals->position.addZ(movZ);
//	warning ("addX:%f", vals->position.X() );
}

void	CameraView::MoveUpDown(float amnt)
{
	vals->position.addY(amnt);
}

//----------------------------------------------------------------
void	CameraView::PanLeftRight(float amnt)	
{	
	vals->rotate.addY(amnt);
	float	deg = 2.0*M_PI*vals->rotate.y()/360.0;
	vals->direction.setX( vals->directionXZ.setX(sin(deg)) );
	vals->direction.setZ( vals->directionXZ.setZ(-cos(deg)) );
}

void	CameraView::PanUpDown(float amnt)	
{	
	vals->rotate.addX(amnt);	
}

//----------------------------------------------------------------
void	CameraView::TrackInOut(float amnt)
{	
	MVertex	nDir(vals->direction);
	nDir.normalize();
	
	AddTransX( -amnt*nDir.x() );
	AddTransY( -amnt*nDir.y() );
	AddTransZ( -amnt*nDir.z() );
}

//----------------------------------------------------------------
void	CameraView::AddOrbitX(float amnt)
{	
	amnt = amnt;
//	float	dist;	// euclidian distance
//	float	deg = 2.0*M_PI*amnt/360.0;
//	float	diffX;
//	float	diffZ;
//	
//	targetDir.set(	vals->target.X()-vals->position.X(),
//					vals->target.Y()-vals->position.Y(),
//					vals->target.Z()-vals->position.Z() );
//	dist = sqrt(targetDir.X()*targetDir.X() + targetDir.Y()*targetDir.Y() + targetDir.Z()*targetDir.Z());
////	targetDir.normalize();
////	vals->direction.normalize();
////	dirDiff = vals->direction - targetDir;
//	
//	diffX = sin(deg)*dist;
//	diffZ = dist - cos(deg)*dist;
//	
//	vals->position.addX(diffX);
//	vals->position.addZ(diffZ);
}
	
//----------------------------------------------------------------
void	CameraView::SetTrans(float x, float y, float z)
{	vals->position.set(x, y, z);}
void	CameraView::SetTrans (MVertex coords)
{	vals->position.set(coords);	}
void	CameraView::SetTransX (float in)
{	vals->position.setX(in);	}
void	CameraView::SetTransY (float in)
{	vals->position.setY(in);	}
void	CameraView::SetTransZ (float in)
{	vals->position.setZ(in);	}

void	CameraView::AddTrans (MVertex coords)
{	vals->position.add(coords);	}
void	CameraView::AddTrans (float x, float y, float z)
{	
	vals->position.addX(x);
	vals->position.addY(y);
	vals->position.addZ(z);
}
void	CameraView::AddTransX (float in)
{	vals->position.addX(in);	}
void	CameraView::AddTransY (float in)
{	vals->position.addY(in);	}
void	CameraView::AddTransZ (float in)
{	vals->position.addZ(in);	}

//----------------------------------------------------------------
void	CameraView::SetRot(float x, float y, float z)
{	vals->rotate.set(x, y, z);}

//----------------------------------------------------------------
void	CameraView::SetTarget(float x, float y, float z)
{	
	vals->target.set(x, y, z);
}

//----------------------------------------------------------------
void	CameraView::Print()
{
	warning("FOV		: %f	Aspect:	%f", vals->fov, vals->aspect);
}

/*---------------------------------------------------------------------------*/
void 	CameraView::SetGridNumUnits(int in)
{
	vals->numUnits = in;
}

/*---------------------------------------------------------------------------*/
void	CameraView::SetGridUnitSize(float in)
{
	vals->unitSize = in;
}

/*---------------------------------------------------------------------------*/
void	CameraView::SetGridIntensity(float in)
{
	vals->intensity = in;
}

/*---------------------------------------------------------------------------*/
void 	CameraView::SetYZGrid(bool status)
{
	vals->gridX = status;
	emit	UpdateRequired();
}

/*---------------------------------------------------------------------------*/
void 	CameraView::SetXZGrid(bool status)
{
	vals->gridY = status;
	emit	UpdateRequired();
}

/*---------------------------------------------------------------------------*/
void 	CameraView::SetXYGrid(bool status)
{
	vals->gridZ = status;
	emit	UpdateRequired();
}

/*---------------------------------------------------------------------------*/
void 	CameraView::DrawGrid()
{
	DrawGlGrid(vals->gridX, vals->gridY, vals->gridZ, vals->unitSize, vals->numUnits, vals->intensity);
}

/*---------------------------------------------------------------------------*/
void	CameraView::SetGrid (bool x, bool y, bool z)
{
	vals->gridX = x;
	vals->gridY = y;
	vals->gridZ = z;
}

//==============================================================================
// CAMERAVALS
//==============================================================================
CameraVals::CameraVals()
{
	projType   = CAM_PERSPECTIVE;
	position.set(0.0, 0.0, 0.0);
	target.set	(0.0, 0.0, 0.0);
	direction.set(0.0, 0.0, 1.0);
	rotate.set	(0.0, 0.0, 0.0);
	fov 	   = 30.0;
	aspect     = 1.0;
	gridX	   = false;
	gridY	   = true;
	gridZ	   = false;
	unitSize   = 1.0;
	numUnits   = 5;
	frame	   = 0;
	targetLock = false;
}

CameraVals::CameraVals(CameraVals &in)
{
	projType   = in.projType;
	position.set	(in.position);
	target.set		(in.target);
	direction.set	(in.direction);
	rotate.set		(in.rotate);
	fov 	   = in.fov;
	aspect     = in.aspect;
	gridX	   = in.gridX;
	gridY	   = in.gridY;
	gridZ	   = in.gridZ;
	unitSize   = in.unitSize;
	numUnits   = in.numUnits;
	frame	   = in.frame;
	targetLock = in.targetLock;
}
CameraVals::~CameraVals()
{}

//--------------------------------------
CameraVals& CameraVals::operator = (CameraVals &in)
{
	projType   = in.projType;
	position.set	(in.position);
	target.set		(in.target);
	direction.set	(in.direction);
	rotate.set		(in.rotate);
	fov 	   = in.fov;
	aspect     = in.aspect;
	gridX	   = in.gridX;
	gridY	   = in.gridY;
	gridZ	   = in.gridZ;
	unitSize   = in.unitSize;
	numUnits   = in.numUnits;
	frame	   = in.frame;
	targetLock = in.targetLock;
	
	return *this;
}

//--------------------------------------
void	CameraVals::PrintValues(FILE* outfile)
{
	char	buffer[128];
	
	sprintf(buffer, "%30s : %d\n", "projType", projType );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "positionX", position.x() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "positionY", position.y() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "positionZ", position.z() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "targetX", target.x() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "targetY", target.y() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "targetZ", target.z() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "directionX", direction.x() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "directionY", direction.y() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "directionZ", direction.z() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "directionXZX", directionXZ.x() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "directionXZY", directionXZ.y() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "directionXZZ", directionXZ.z() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "rotateX", rotate.x() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "rotateY", rotate.y() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "rotateZ", rotate.z() );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "fov", fov );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "aspect", aspect );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %d\n", "gridX", gridX );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %d\n", "gridY", gridY );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %d\n", "gridZ", gridZ );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %f\n", "unitSize", unitSize );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %d\n", "numUnits", numUnits );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %d\n", "frame", frame );
	fputs(buffer, outfile);
	sprintf(buffer, "%30s : %d\n", "targetLock", targetLock );
	fputs(buffer, outfile);
	sprintf(buffer, " \n");
	fputs(buffer, outfile);
}




