/**
 * configFileIO.cpp
 * Copyright 1999 by Mark B. Allan
 * 
 * Generic config file helper classes
 */
#include "configFileIO.h"

#include <stdio.h>

/**
 * read config file
 */
//==============================================================================
ConfigFileReader::ConfigFileReader(const char* filename)
{
	isValid = false;
	
	configDict = new QAsciiDict<char>(101, true, true);
	configDict->setAutoDelete(true);
	char	*deepCopy;
	
	if (!filename)
		return;
		
	FILE	*infile;
	infile = fopen(filename, "r");
	if (!infile)
	{
		warning("ConfigFileReader::ConfigFileReader(%s): file not found. No big deal.", filename);
		return;
	}
	
	int		i;
	char	buffer[256];
	QCString	strKey;
	QCString	strValue;
	while( fgets(buffer, 256, infile) )
	{
		strKey = buffer;
		strKey = strKey.simplifyWhiteSpace();
		if ( (i = strKey.find("#")) == 0)
			continue;
		else if ( i > 0 )
			strKey.truncate(i);
		else if ( strKey.length() < 3 )
			continue;
			
		strValue = (const char*)strKey;
		i = strKey.find(" ");
		if ( i < 1 )
			continue;
		strKey.truncate(i);
		strValue.remove(0, i+1);
		
		deepCopy = new char[strValue.length()+1];
		strcpy(deepCopy, (const char*)strValue);
		configDict->insert( (const char*)strKey, deepCopy );
		//warning("key = %-20s, value = %-20s", (const char*)strKey, deepCopy );
	}
	isValid = true;
}

ConfigFileReader::~ConfigFileReader()
{
}

/**
 * find boolean value based on key
 * @return success
 */
//--------------------------------------------------------------------
bool	ConfigFileReader::findBool(const char* key, bool& value)
{
	QCString	strValue;
	bool	retVal = false;
	if(configDict->find(key))
	{
		strValue = configDict->find(key);
		if(strValue.find('f') == 0 || strValue.find('F') == 0)
		{
			value = false;
			retVal = true;
		}
		else if (strValue.find('t') == 0 || strValue.find('T') == 0)
		{
			value = true;
			retVal = true;
		}
	}
	return retVal;
}

/**
 * find integer value based on key
 * @return success
 */
//--------------------------------------------------------------------
bool	ConfigFileReader::findInt(const char* key, int& value)
{
	bool	retVal = false;
	int		tmp;
	QCString strValue;
	if(configDict->find(key))
	{
		strValue = configDict->find(key);
		tmp = strValue.toInt(&retVal);
		if(retVal)
			value = tmp;
	}
	return retVal;
}

/**
 * find float value based on key
 * @return success
 */
//--------------------------------------------------------------------
bool	ConfigFileReader::findFloat(const char* key, float& value)
{
	bool	retVal = false;
	float	tmp;
	QCString strValue;
	if(configDict->find(key))
	{
		strValue = configDict->find(key);
		tmp = strValue.toFloat(&retVal);
		if(retVal)
			value = tmp;
	}
	return retVal;
}

/**
 * find string value based on key
 * @return success
 */
//--------------------------------------------------------------------
bool	ConfigFileReader::findString(const char* key, char*& value)
{
	bool	retVal = false;
	QCString strValue;
	if(configDict->find(key))
	{
		strValue = configDict->find(key);
		if(!strValue.isEmpty())
		{
			if(value != 0)
				delete [] value;
			value = new char[strValue.length()+1];
			strcpy(value,(const char*)strValue);
			retVal = true;
		}
	}
	return retVal;
}

//--------------------------------------------------------------------
bool	ConfigFileReader::findToolBarDock(const char* key, int& dock)
{
	bool	retVal = false;
	QCString strValue;
	if(configDict->find(key))
	{
		strValue = configDict->find(key);
		dock = (int)ToolBarValue(strValue);
		retVal = true;
	}
	return retVal;
}

//--------------------------------------------------------------------
QMainWindow::ToolBarDock	ConfigFileReader::ToolBarValue(QCString& strValue)
{
	if ( strValue.find('t') == 0 || strValue.find('T') == 0 )
		return  QMainWindow::Top;
	else if ( strValue.find('b') == 0 || strValue.find('B') == 0 )
		return  QMainWindow::Bottom;
	else if ( strValue.find('l') == 0 || strValue.find('L') == 0 )
		return  QMainWindow::Left;
	else if ( strValue.find('r') == 0 || strValue.find('R') == 0 )
		return  QMainWindow::Right;
	else
		return  QMainWindow::Top;
}

/**
 * write config file
 */
//==============================================================================
ConfigFileWriter::ConfigFileWriter(const char* filename)
{
	isValid = false;
	
	if (!filename)
		return;
		
	outfile = fopen(filename, "w");
	if (!outfile)
	{
		warning("ConfigFileWriter::ConfigFileWriter(%s): COULD NOT OPEN FOR WRITE", filename);
		return;
	}
	isValid = true;
}

ConfigFileWriter::~ConfigFileWriter()
{
	if(outfile)
		fclose(outfile);
}

//--------------------------------------------------------------------
void	ConfigFileWriter::writeBool(const char* key, bool value)
{
	if(outfile)
	{
		if(value)
			sprintf(buffer, "%-23s true\n", key);
		else
			sprintf(buffer, "%-23s false\n", key);
		fputs(buffer, outfile);
	}
}

//--------------------------------------------------------------------
void	ConfigFileWriter::writeInt(const char* key, int value)
{
	if(outfile)
	{
		sprintf(buffer, "%-23s %d\n", key, value);
		fputs(buffer, outfile);
	}
}

//--------------------------------------------------------------------
void	ConfigFileWriter::writeFloat(const char* key, float value)
{
	if(outfile)
	{
		sprintf(buffer, "%-23s %f\n", key, value);
		fputs(buffer, outfile);
	}
}

//--------------------------------------------------------------------
void	ConfigFileWriter::writeString(const char* key, char* value)
{
	if(outfile)
	{
		sprintf(buffer, "%-23s %s\n", key, value);
		fputs(buffer, outfile);
	}
}

//--------------------------------------------------------------------
void	ConfigFileWriter::writeToolBarDock(const char* key, int dock)
{
	if(outfile)
	{
		sprintf(buffer, "%-23s %s\n", key, ToolBarString(dock));
		fputs(buffer, outfile);
	}
}

//--------------------------------------------------------------------
void	ConfigFileWriter::writeComment(const char* comment)
{
	if(outfile)
	{
		sprintf(buffer, "#%s\n", comment);
		fputs(buffer, outfile);
	}
}

//--------------------------------------------------------------------
const char*	ConfigFileWriter::ToolBarString( int tbDock)
{
	switch((QMainWindow::ToolBarDock)tbDock)
	{
		case QMainWindow::Top:
			return "top";
		case QMainWindow::Bottom:
			return "bottom";
		case QMainWindow::Left:
			return "left";
		case QMainWindow::Right:
			return "right";
		case QMainWindow::Unmanaged:
			return "top";
		case QMainWindow::TornOff:
			return "top";
		default:
			return "top";
	}
}


	






