/**
 * glDomain.cpp
 * Copyright 1999 by Mark B. Allan
 * 
 * gl representation of a domain
 */
#include "glDomain.h"

#include "extern.h"
#include "basetypes.h"
#include "mmath.h"
#include "glPrimitives.h"


//==============================================================================
GLDomain::GLDomain()
{
	displayList = 0;
}

GLDomain::~GLDomain()
{
	if ( glIsList(displayList) )
	{
		//warning("delete gl list");
		glDeleteLists(displayList, 1);
	}
}

//--------------------------------------------------------------------
void GLDomain::CreateListFromDomain(DrawStyle dS, PDomain &domain, int detail, float depth)
{
	if ( glIsList(displayList) )
	{
		glDeleteLists(displayList, 1);
	}

	displayList = glGenLists(1);
	if(displayList)
	{
		glNewList(displayList, GL_COMPILE);
		switch(domain.Type())
		{
			case PDPoint:
				DrawPointGL(dS, domain);
				break;
			case PDLine:  
				DrawLineGL(dS, domain);
				break;
			case PDTriangle:
				DrawTriangleGL(dS, domain, detail, depth);
				break;
			case PDPlane:
				DrawPlaneGL(dS, domain, detail, depth); 
				break;
			case PDBox:
				DrawBoxGL(dS, domain);
				break;
			case PDSphere:
				DrawSphereGL(dS, domain, detail);
				break;
			case PDCylinder:
				DrawCylinderGL(dS, domain);
				break;
			case PDCone:
				DrawConeGL(dS, domain);
				break;
			case PDBlob:
				DrawBlobGL(dS, domain);
				break;
			case PDDisc:
				DrawDiskGL(dS, domain, detail, depth);
				break;
		}
		glEndList();
	}
	else
		warning("GLDomain::CreateListFromDomain - displayList == 0");
	
}


//--------------------------------------------------------------------
void	GLDomain::DrawPointGL(DrawStyle, PDomain& domain)
{
	if(g_glThick) glPointSize(5.0);
		
	glNormal3f(0.0, 1.0, 0.0);
	glBegin(GL_POINTS);
		glVertex3f( domain.A0(), domain.A1(), domain.A2() );
	glEnd();
}
//--------------------------------------------------------------------
void	GLDomain::DrawLineGL(DrawStyle, PDomain& domain)
{
	if(g_glThick) glLineWidth(3.0);
	glNormal3f(0.0, 1.0, 0.0);
	glBegin(GL_LINES);
		glVertex3f( domain.A0(), domain.A1(), domain.A2() );
		glVertex3f( domain.A3(), domain.A4(), domain.A5() );
	glEnd();
}
//--------------------------------------------------------------------
void	GLDomain::DrawTriangleGL(DrawStyle dS, PDomain& domain, float detail, float depth)
{
	// This needs work. How to handle clockwise/counterclockwise &
	// how to handle normal?
	detail = detail;
	depth = depth;
	
	float	vert1[3] = { domain.A0(), domain.A1(), domain.A2() };
	float	vert2[3] = { domain.A3(), domain.A4(), domain.A5() };
	float	vert3[3] = { domain.A6(), domain.A7(), domain.A8() };
	float	norm[3];
	
	::calcNormal(norm, vert1, vert2, vert3);
	
	switch(dS)
	{
		case Filled:
			glNormal3fv(norm);
			glBegin(GL_TRIANGLES);
				glVertex3f( domain.A0(), domain.A1(), domain.A2() );
				glVertex3f( domain.A3(), domain.A4(), domain.A5() );
				glVertex3f( domain.A6(), domain.A7(), domain.A8() );
			glEnd();
			break;
		case Wire:
			glNormal3f(0.0, 1.0, 0.0);
			glLineWidth(1);
			glBegin(GL_LINE_LOOP);
				glVertex3f( domain.A0(), domain.A1(), domain.A2() );
				glVertex3f( domain.A3(), domain.A4(), domain.A5() );
				glVertex3f( domain.A6(), domain.A7(), domain.A8() );
			glEnd();
			break;
	}
		
}

//--------------------------------------------------------------------
void	GLDomain::DrawPlaneGL(DrawStyle dS, PDomain& domain, float detail, float depth)
{
	float	emis[4] = { 1.00, 1.00, 1.00, 1.00 };
	float	A0[3] = {	domain.A0(), 
						domain.A1(), 
						domain.A2() };
	float	B0[3] = {	A0[0] + domain.A6(),
					 	A0[1] + domain.A7(),
					 	A0[2] + domain.A8() };
	float	C0[3] = {	B0[0] + domain.A3(),
						B0[1] + domain.A4(),
						B0[2] + domain.A5() };
	float	D0[3] = {	A0[0] + domain.A3(), 
						A0[1] + domain.A4(), 
						A0[2] + domain.A5() };

	float	p0[2][2][3];
	
	glLineWidth(1.0);
	switch(dS)
	{
		case Filled:
			// Get corners A, B, C and D
			p0[0][0][0] = C0[0];	p0[0][0][1] = C0[1];	p0[0][0][2] = -C0[2];
			p0[0][1][0] = B0[0];	p0[0][1][1] = B0[1];	p0[0][1][2] = -B0[2];
			p0[1][1][0] = A0[0];	p0[1][1][1] = A0[1];	p0[1][1][2] = -A0[2];
			p0[1][0][0] = D0[0];	p0[1][0][1] = D0[1];	p0[1][0][2] = -D0[2];

			RglBilinearPatch(p0, (int)detail);
			//-- if plane has depth, draw it...
			if (depth != 0.0f)
			{
				MVertex	v1(D0[0], D0[1], D0[2]);
				MVertex v2(A0[0], A0[1], A0[2]);
				MVertex v3(B0[0], B0[1], B0[2]);
				MVertex normal;
				normal.CreateNormal(v1, v2, v3);
				normal.setX(normal.x()*depth);
				normal.setY(normal.y()*depth);
				normal.setZ(normal.z()*depth);

				float	A1[3];
				float	B1[3];
				float	C1[3];
				float	D1[3];
				A1[0] = A0[0]-normal.x();
				A1[1] = A0[1]-normal.y();
				A1[2] = A0[2]-normal.z();
				B1[0] = B0[0]-normal.x();
				B1[1] = B0[1]-normal.y();
				B1[2] = B0[2]-normal.z();
				C1[0] = C0[0]-normal.x();
				C1[1] = C0[1]-normal.y();
				C1[2] = C0[2]-normal.z();
				D1[0] = D0[0]-normal.x();
				D1[1] = D0[1]-normal.y();
				D1[2] = D0[2]-normal.z();

				p0[0][0][0] = C1[0];	p0[0][0][1] = C1[1];	p0[0][0][2] = -C1[2];
				p0[0][1][0] = D1[0];	p0[0][1][1] = D1[1];	p0[0][1][2] = -D1[2];
				p0[1][1][0] = A1[0];	p0[1][1][1] = A1[1];	p0[1][1][2] = -A1[2];
				p0[1][0][0] = B1[0];	p0[1][0][1] = B1[1];	p0[1][0][2] = -B1[2];
				float	p1[2][2][3];
				float	p2[2][2][3];
				float	p3[2][2][3];
				float	p4[2][2][3];
				p1[0][0][0] = B0[0];	p1[0][0][1] = B0[1];	p1[0][0][2] = -B0[2];
				p1[0][1][0] = B1[0];	p1[0][1][1] = B1[1];	p1[0][1][2] = -B1[2];
				p1[1][1][0] = A1[0];	p1[1][1][1] = A1[1];	p1[1][1][2] = -A1[2];
				p1[1][0][0] = A0[0];	p1[1][0][1] = A0[1];	p1[1][0][2] = -A0[2];
				p2[0][0][0] = C0[0];	p2[0][0][1] = C0[1];	p2[0][0][2] = -C0[2];
				p2[0][1][0] = C1[0];	p2[0][1][1] = C1[1];	p2[0][1][2] = -C1[2];
				p2[1][1][0] = B1[0];	p2[1][1][1] = B1[1];	p2[1][1][2] = -B1[2];
				p2[1][0][0] = B0[0];	p2[1][0][1] = B0[1];	p2[1][0][2] = -B0[2];
				p3[0][0][0] = D0[0];	p3[0][0][1] = D0[1];	p3[0][0][2] = -D0[2];
				p3[0][1][0] = D1[0];	p3[0][1][1] = D1[1];	p3[0][1][2] = -D1[2];
				p3[1][1][0] = C1[0];	p3[1][1][1] = C1[1];	p3[1][1][2] = -C1[2];
				p3[1][0][0] = C0[0];	p3[1][0][1] = C0[1];	p3[1][0][2] = -C0[2];
				p4[0][0][0] = A0[0];	p4[0][0][1] = A0[1];	p4[0][0][2] = -A0[2];
				p4[0][1][0] = A1[0];	p4[0][1][1] = A1[1];	p4[0][1][2] = -A1[2];
				p4[1][1][0] = D1[0];	p4[1][1][1] = D1[1];	p4[1][1][2] = -D1[2];
				p4[1][0][0] = D0[0];	p4[1][0][1] = D0[1];	p4[1][0][2] = -D0[2];

				int	sideDetail = (int)detail/2;
				RglBilinearPatch(p0, (int)detail);
				RglBilinearPatch(p1, sideDetail);
				RglBilinearPatch(p2, sideDetail);
				RglBilinearPatch(p3, sideDetail);
				RglBilinearPatch(p4, sideDetail);
			}
			break;
		case Wire:
			glNormal3f( 0.0, 1.0, 0.0);
			glLineWidth(1);
			glBegin(GL_LINE_LOOP);
				glVertex3f (	domain.A0(), 
								domain.A1(), 
								domain.A2() );
				glVertex3f (	domain.A0()+domain.A3(),
								domain.A1()+domain.A4(),
								domain.A2()+domain.A5() );
				glVertex3f (	domain.A0()+domain.A3()+domain.A6(),
								domain.A1()+domain.A4()+domain.A7(),
								domain.A2()+domain.A5()+domain.A8() );
				glVertex3f (	domain.A0()+domain.A6(),
								domain.A1()+domain.A7(),
								domain.A2()+domain.A8() );
			glEnd();
			break;
	}
	//-- draw corner for reference point...
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION,	emis);
	if(g_glThick) glPointSize(5.0);
	glBegin(GL_POINTS);
		glVertex3f(domain.A0(), domain.A1(), domain.A2() );
	glEnd();
	
}

//--------------------------------------------------------------------
void	GLDomain::DrawBoxGL(DrawStyle dS, PDomain& domain )
{
	switch(dS)
	{
		case Filled:
		case Wire:
			glNormal3f(0.0, 1.0, 0.0);
			glLineWidth(1.0);
			glBegin(GL_LINE_LOOP);
				glVertex3f(domain.A0(), domain.A1(), domain.A2());
				glVertex3f(domain.A0(), domain.A4(), domain.A2());
				glVertex3f(domain.A3(), domain.A4(), domain.A2());
				glVertex3f(domain.A3(), domain.A1(), domain.A2());
				glVertex3f(domain.A3(), domain.A1(), domain.A5());
				glVertex3f(domain.A3(), domain.A4(), domain.A5());
				glVertex3f(domain.A0(), domain.A4(), domain.A5());
				glVertex3f(domain.A0(), domain.A1(), domain.A5());
				glVertex3f(domain.A0(), domain.A1(), domain.A2());
				glVertex3f(domain.A3(), domain.A1(), domain.A2());
				glVertex3f(domain.A3(), domain.A4(), domain.A2());
				glVertex3f(domain.A3(), domain.A4(), domain.A5());
				glVertex3f(domain.A3(), domain.A1(), domain.A5());
				glVertex3f(domain.A0(), domain.A1(), domain.A5());
				glVertex3f(domain.A0(), domain.A4(), domain.A5());
				glVertex3f(domain.A0(), domain.A4(), domain.A2());
			glEnd();
	}

}

//--------------------------------------------------------------------
void	GLDomain::DrawSphereGL(DrawStyle dS, PDomain& domain, float detail)
{
	#define	CSTEPS	24
	int i;
	static	bool	once = true;
	static	float	circle[CSTEPS][2];
	if(once)
	{
		once = false;
		float	rinc = (2.0*M_PI)/((double)(CSTEPS));
		for (i = 0; i < CSTEPS; i++)
		{
			circle[i][0] = sin((float)i*rinc);
			circle[i][1] = cos((float)i*rinc);
		}
	}
	
	float	X = domain.A0();
	float	Y = domain.A1();
	float	Z = domain.A2();
	float	r1 = domain.A3();
	float	r2 = domain.A4();
	
	switch(dS)
	{
		case Filled:
			glPushMatrix();
			glTranslatef( X, Y, Z );
			RglSphere( r1, -r1, r1, 360.0, (int)detail );
			glPopMatrix();
			break;
		case Wire:
			glNormal3f(0.0, 1.0, 0.0);
			glLineWidth(1);
			if (r2 <= 0.0)
				r2 = r1;
			glBegin(GL_LINE_LOOP);
				for (i = 0; i < CSTEPS; i++)
					glVertex3f( (circle[i][0]*r1)+X, (circle[i][1]*r1)+Y, Z);
			glEnd();
			glBegin(GL_LINE_LOOP);
				for (i = 0; i < CSTEPS; i++)
					glVertex3f( (circle[i][0]*r1)+X, Y, (circle[i][1]*r1)+Z);
			glEnd();
			glBegin(GL_LINE_LOOP);
				for (i = 0; i < CSTEPS; i++)
					glVertex3f( X, (circle[i][0]*r1)+Y, (circle[i][1]*r1)+Z);
			glEnd();
			
			if (r2 > 0.0)
			{
			glBegin(GL_LINES);
				for (i = 0; i < CSTEPS; i++)
					glVertex3f( (circle[i][0]*r2)+X, (circle[i][1]*r2)+Y, Z);
			glEnd();
			glBegin(GL_LINES);
				for (i = 0; i < CSTEPS; i++)
					glVertex3f( (circle[i][0]*r2)+X, Y, (circle[i][1]*r2)+Z);
			glEnd();
			glBegin(GL_LINES);
				for (i = 0; i < CSTEPS; i++)
					glVertex3f( X, (circle[i][0]*r2)+Y, (circle[i][1]*r2)+Z);
			glEnd();
			
			}
			break;
	}
	#undef CSTEPS
}
//--------------------------------------------------------------------
void	GLDomain::DrawCylinderGL(DrawStyle dS, PDomain& domain )
{
	#define	CSTEPS	12
//	int i;
//	static	bool	once = true;
//	static	float	circle[CSTEPS][2];
//	if(once)
//	{
//		once = false;
//		float	rinc = (2.0*M_PI)/((double)(CSTEPS));
//		for (i = 0; i < CSTEPS; i++)
//		{
//			circle[i][0] = sin((float)i*rinc);
//			circle[i][1] = cos((float)i*rinc);
//		}
//	}
	
	MVertex	pos1(domain.A0(), domain.A1(), domain.A2());
	MVertex	pos2(domain.A3(), domain.A4(), domain.A5());
	switch(dS)
	{
		case Filled:
		case Wire:
			glLineWidth(1);
			glBegin(GL_LINES);
				glVertex3fv(pos1.XYZ());
				glVertex3fv(pos2.XYZ());
			glEnd();
			break;
	}
	#undef CSTEPS
}
//--------------------------------------------------------------------
void	GLDomain::DrawConeGL(DrawStyle dS, PDomain& domain )
{
	#define	CSTEPS	12
//	int i;
//	static	bool	once = true;
//	static	float	circle[CSTEPS][2];
//	if(once)
//	{
//		once = false;
//		float	rinc = (2.0*M_PI)/((double)(CSTEPS));
//		for (i = 0; i < CSTEPS; i++)
//		{
//			circle[i][0] = sin((float)i*rinc);
//			circle[i][1] = cos((float)i*rinc);
//		}
//	}
	
	MVertex	pos1(domain.A0(), domain.A1(), domain.A2());
	MVertex	pos2(domain.A3(), domain.A4(), domain.A5());
	switch(dS)
	{
		case Filled:
		case Wire:
			glLineWidth(1);
			glBegin(GL_LINES);
				glVertex3fv(pos1.XYZ());
				glVertex3fv(pos2.XYZ());
			glEnd();
			break;
	}
	#undef CSTEPS
}

//--------------------------------------------------------------------
void	GLDomain::DrawBlobGL(DrawStyle dS, PDomain& domain )
{	
	#define	CSTEPS	100
	int i;
	static	bool	once = true;
	static	float	circle[CSTEPS][2];
	if(once)
	{
		once = false;
		float	rinc = (2.0*M_PI)/((double)(CSTEPS));
		for (i = 0; i < CSTEPS; i++)
		{
			circle[i][0] = sin((float)i*rinc);
			circle[i][1] = cos((float)i*rinc);
		}
	}
	
	
	float	X = domain.A0();
	float	Y = domain.A1();
	float	Z = domain.A2();
	float	r1= domain.A3()*2.0;
	switch(dS)
	{
		case Filled:
		case Wire:
			glNormal3f(0.0, 1.0, 0.0);
			if(g_glThick)	glPointSize(5.0);
			glBegin(GL_POINTS);
				glVertex3f(X, Y, Z);
			glEnd();
			glLineWidth(1.0);
			glBegin(GL_LINES);
				for (i = 0; i < CSTEPS-1; i+=5)
				{
					glVertex3f( (circle[i  ][0]*r1)+X, (circle[i  ][1]*r1)+Y, Z);
					glVertex3f( (circle[i+1][0]*r1)+X, (circle[i+1][1]*r1)+Y, Z);
					
					glVertex3f( (circle[i  ][0]*r1)+X, Y, (circle[i  ][1]*r1)+Z);
					glVertex3f( (circle[i+1][0]*r1)+X, Y, (circle[i+1][1]*r1)+Z);
					
					glVertex3f( X, (circle[i  ][0]*r1)+Y, (circle[i  ][1]*r1)+Z);
					glVertex3f( X, (circle[i+1][0]*r1)+Y, (circle[i+1][1]*r1)+Z);
				}
			glEnd();
			break;
	}
	#undef CSTEPS
}

//--------------------------------------------------------------------
void	GLDomain::DrawDiskGL(DrawStyle dS, PDomain& domain, float detail, float depth )
{
	MMatrix	rot;
	
	switch(dS)
	{
		case Filled:
		case Wire:
			rot.pointAt(MVertex(domain.A3(), domain.A4(), domain.A5()));
			rot.setTranslate(domain.A0(), domain.A1(), domain.A2());
			glPushMatrix();
				glMultMatrixf( rot.mat() );
				RglDisk(0.0, domain.A6(), 360.0, (int)detail);
				if(depth)
				{
					RglCylinder(domain.A6(), -depth, 0.0, 360.0, (int)detail);
				}
			glPopMatrix();
			break;
	}
//	#define	CSTEPS	100
//	int i;
//	static	bool	once = true;
//	static	float	circle[CSTEPS][2];
//	if(once)
//	{
//		once = false;
//		float	rinc = (2.0*M_PI)/((double)(CSTEPS));
//		for (i = 0; i < CSTEPS; i++)
//		{
//			circle[i][0] = sin((float)i*rinc);
//			circle[i][1] = cos((float)i*rinc);
//		}
//	}
//	
//	
//	float	X = domain.A0();
//	float	Y = domain.A1();
//	float	Z = domain.A2();
//	float	r1= domain.A3()*2.0;
//	switch(dS)
//	{
//		case Filled:
//		case Wire:
//			glNormal3f(0.0, 1.0, 0.0);
//			glPointSize(5);
//			glBegin(GL_POINTS);
//				glVertex3f(X, Y, Z);
//			glEnd();
//			glLineWidth(1);
//			glBegin(GL_LINES);
//				for (i = 0; i < CSTEPS-1; i+=5)
//				{
//					glVertex3f( (circle[i  ][0]*r1)+X, (circle[i  ][1]*r1)+Y, Z);
//					glVertex3f( (circle[i+1][0]*r1)+X, (circle[i+1][1]*r1)+Y, Z);
//					
//					glVertex3f( (circle[i  ][0]*r1)+X, Y, (circle[i  ][1]*r1)+Z);
//					glVertex3f( (circle[i+1][0]*r1)+X, Y, (circle[i+1][1]*r1)+Z);
//					
//					glVertex3f( X, (circle[i  ][0]*r1)+Y, (circle[i  ][1]*r1)+Z);
//					glVertex3f( X, (circle[i+1][0]*r1)+Y, (circle[i+1][1]*r1)+Z);
//				}
//			glEnd();
//			break;
//	}
//	#undef CSTEPS
}

//--------------------------------------------------------------------
void	GLDomain::CreateListRMSphere(MMatrix &trans, float radius, float zmin, float zmax, float thetamax, float detail, bool closeTheta, bool flatCap)
{
	if ( glIsList(displayList) )
	{
		glDeleteLists(displayList, 1);
	}
		
	displayList = glGenLists(1);
	glNewList(displayList, GL_COMPILE);
		glPushMatrix();
		glMultMatrixf( trans.mat() );
		if (closeTheta)
		{
		}
		if (flatCap)
		{
			float cap;
			if ( fabs(zmin) < fabs(radius) )
			{
				cap = sqrt(radius*radius - zmin*zmin);
				RglDisk(zmin, cap, thetamax, (int)detail );
			}
			if ( fabs(zmax) < fabs(radius) )
			{
				cap = sqrt(radius*radius - zmax*zmax);
				RglDisk(zmax, cap, thetamax, (int)detail );
			}
		}
		RglSphere( radius, zmin, zmax, thetamax, (int)detail );
		glPopMatrix();
	glEndList();
}

//--------------------------------------------------------------------
void	GLDomain::CreateListRMCone(MMatrix &trans, float height, float radius, float thetamax, float detail, bool closeTheta, bool flatCap, bool roundCap)
{
	if ( glIsList(displayList) )
	{
		glDeleteLists(displayList, 1);
	}
		
	displayList = glGenLists(1);
	glNewList(displayList, GL_COMPILE);
		glPushMatrix();
		glMultMatrixf( trans.mat() );
		if (flatCap)
		{
			RglDisk(0.0, radius, thetamax, (int)detail);
		}
		if (roundCap)
		{
			RglSphere(radius, -radius, 0.0f, thetamax, (int)detail);
		}
		if (closeTheta)
		{
			if(roundCap)
			{
			}
		}
		RglCone( height, radius, thetamax, (int)detail );
		glPopMatrix();
	glEndList();
}

//--------------------------------------------------------------------
void	GLDomain::CreateListRMDisk(MMatrix &trans, float height, float radius, float thetamax, float detail)
{
	if ( glIsList(displayList) )
	{
		glDeleteLists(displayList, 1);
	}
		
	displayList = glGenLists(1);
	glNewList(displayList, GL_COMPILE);
		glPushMatrix();
		glMultMatrixf( trans.mat() );
		RglDisk( height, radius, thetamax, (int)detail );
		glPopMatrix();
	glEndList();
}

//--------------------------------------------------------------------
void	GLDomain::CreateListRMCylinder(	MMatrix &trans, float radius, float zmin, float zmax, float thetamax, 
										float detail, bool closeTheta, bool flatCap, bool roundCap)
{
	if ( glIsList(displayList) )
	{
		glDeleteLists(displayList, 1);
	}
		
	displayList = glGenLists(1);
	glNewList(displayList, GL_COMPILE);
		glPushMatrix();
		glMultMatrixf( trans.mat() );
		if (flatCap)
		{
			RglDisk(zmin, radius, thetamax, (int)detail);
			RglDisk(zmax, radius, thetamax, (int)detail);
		}
		if (roundCap)
		{
			glTranslatef(0.0f, 0.0f, zmin);
			RglSphere(radius, -radius, 0.0f, thetamax, (int)detail);
			glTranslatef(0.0f, 0.0f, -zmin+zmax);
			RglSphere(radius, 0.0f,  radius, thetamax, (int)detail);
			glTranslatef(0.0f, 0.0f, -zmax);
		}
		if (closeTheta)
		{
			if(roundCap)
			{
			}
		}
		RglCylinder( radius, zmin, zmax, thetamax, (int)detail );
		glPopMatrix();
	glEndList();
}


