#include "hitList.h"

#include "compatibility.h"
#include <GL/gl.h>
#include "extern.h"
#include "glPortCam.h"
#include "mbaLight.h"
#include "pAction.h"
#include "particleControlWidget.h"
#include "lightGroupWidget.h"
#include "camera.h"

//==============================================================================
HitPrimitive::HitPrimitive(PrimType pType, void* pTarget, const float *in, const float *h1, const float *h2)
{
	primType	= pType;
	primTarget	= pTarget;
	hasHit1		= false;
	hasHit2		= false;
	if(in)
	{
		pos[0] = in[0];
		pos[1] = in[1];
		pos[2] = in[2];
	}
	else
	{
		pos[0] = 0.0;
		pos[1] = 0.0;
		pos[2] = 0.0;
	}
	if(h1)
	{
		hasHit1 = true;
		hit1[0] = h1[0];
		hit1[1] = h1[1];
		hit1[2] = h1[2];
	}
	if(h2)
	{
		hasHit1 = true;
		hit1[0] = h2[0];
		hit1[1] = h2[1];
		hit1[2] = h2[2];
	}
}

//==============================================================================
/**
 * list of 'hittable' objects for ortho picking
 */
HitList::HitList()
{
	primitiveList.setAutoDelete(true);
	activePrimitive = 0;
	hitBox = 0.05;
	fov = 1.0;
	autoUpdate = true;
	hilightHitZones = false;
}

HitList::~HitList()
{
}

//--------------------------------------------------------------------
void	HitList::reset()
{
	if(activePrimitive)
		oldActivePrimitiveTarget = activePrimitive->primTarget;
	else
		oldActivePrimitiveTarget = 0;
	primitiveList.clear();
	activePrimitive = 0;
}

//--------------------------------------------------------------------
void	HitList::checkForOldActiveTarget()
{
	int i;
	int count = (int)primitiveList.count();
	for(i = 0; i < count; i++)
	{
//		warning("%d : %p %p", i, primitiveList.at(i)->primTarget, oldActivePrimitiveTarget);
		if(primitiveList.at(i)->primTarget == oldActivePrimitiveTarget)
		{
			activePrimitive = primitiveList.at(i);
			//break;
		}
	}	
}

//--------------------------------------------------------------------
void	HitList::addLightPrimitive(MBALight *l)
{
	HitPrimitive	*tmp = 0;
	if(l->LightType() == MBALight::Local)
		tmp = new HitPrimitive(HitPrimitive::Light, l, l->getFrom());
	else
		tmp = new HitPrimitive(HitPrimitive::Light, l, l->getFrom(), l->getTo());
	
	if(tmp)
		primitiveList.append(tmp);
}

//--------------------------------------------------------------------
void	HitList::addCameraPrimitive(CameraView *cv)
{
	HitPrimitive	*tmp = 0;
	tmp = new HitPrimitive(HitPrimitive::Camera, cv, cv->GetVals().position.XYZ(), cv->GetVals().target.XYZ() );
	primitiveList.append(tmp);
}

//--------------------------------------------------------------------
void	HitList::addActionPrimitive(FlowParticleAction *pa)
{
	HitPrimitive	*tmp = 0;
	float	pos[3];	
	FlowParticleAction::PAType	type = pa->Type();
	if(	type == FlowParticleAction::PA_Bounce || 
		type == FlowParticleAction::PA_Sink || 
		type == FlowParticleAction::PA_Source)
	{
			pos[0] = pa->Domain(0)->A0();
			pos[1] = pa->Domain(0)->A1();
			pos[2] = pa->Domain(0)->A2();
			tmp = new HitPrimitive(HitPrimitive::Action, pa, pos);
			setActionDomainHits(pa, tmp);
	}
			
	if(tmp)
		primitiveList.append(tmp);
}

//--------------------------------------------------------------------
void	HitList::moveActive(float x, float y, float z, bool gridSnap, float gridSnapMult)
{
	float	*moved;
	float	*tmp;
	if(activePrimitive)
	{
		if(activePrimitive->wasHit == 0)
		{
			moved = activePrimitive->pos;
			moved[0] += x;
			moved[1] += y;
			moved[2] += z;
			switch(activePrimitive->primType)
			{
				case HitPrimitive::Undefined:
					break;
				case HitPrimitive::Light:
					moveLightFrom(moved, gridSnap, gridSnapMult);
					break;
				case HitPrimitive::Action:
					moveActionDomainPos(moved, gridSnap, gridSnapMult);
					break;
				case HitPrimitive::Camera:
					if(gridSnap)
					{
						makeSnapCoords(moved, gridSnapMult);
						tmp = snap;
					}
					else
						tmp = moved;
					((CameraView*)(activePrimitive->primTarget))->SetTrans(tmp[0], tmp[1], tmp[2]);
					break;
			}
		}
		else if(activePrimitive->wasHit == 1)
		{
			moved = activePrimitive->hit1;
			moved[0] += x;
			moved[1] += y;
			moved[2] += z;
			switch(activePrimitive->primType)
			{
				case HitPrimitive::Undefined:
					break;
				case HitPrimitive::Light:
					moveLightTo (moved, gridSnap, gridSnapMult);
					//((MBALight*)(activePrimitive->primTarget))->setTo(moved[0], moved[1], moved[2]);
					break;
				case HitPrimitive::Action:
					moveActionDomainHit1(moved, gridSnap, gridSnapMult);
					break;
				case HitPrimitive::Camera:
					if(gridSnap)
					{
						makeSnapCoords(moved, gridSnapMult);
						tmp = snap;
					}
					else
						tmp = moved;
					((CameraView*)(activePrimitive->primTarget))->SetTarget(tmp[0], tmp[1], tmp[2]);
					break;
			}
		}
		else if(activePrimitive->wasHit == 2)
		{
			moved = activePrimitive->hit2;
			moved[0] += x;
			moved[1] += y;
			moved[2] += z;
			switch(activePrimitive->primType)
			{
				case HitPrimitive::Undefined:
					break;
				case HitPrimitive::Light:
					break;
				case HitPrimitive::Action:
					moveActionDomainHit2(moved, gridSnap, gridSnapMult);
					break;
				case HitPrimitive::Camera:
					break;
			}
		}
		if(autoUpdate)
			g_camGLPort->updateGLIfNecessary();
	}
}

//--------------------------------------------------------------------
void	HitList::makeSnapCoords(float *moved, float gridSnapMult)
{
	float	grid = g_gridUnitSize*gridSnapMult;
	float	halfGrid = grid*0.5;
	if(moved[0] > 0)	snap[0] = moved[0]+halfGrid-fmod((moved[0]+halfGrid), grid);
	else				snap[0] = moved[0]-halfGrid-fmod((moved[0]-halfGrid), grid);
	if(moved[1] > 0)	snap[1] = moved[1]+halfGrid-fmod((moved[1]+halfGrid), grid);
	else				snap[1] = moved[1]-halfGrid-fmod((moved[1]-halfGrid), grid);
	if(moved[2] > 0)	snap[2] = moved[2]+halfGrid-fmod((moved[2]+halfGrid), grid);
	else				snap[2] = moved[2]-halfGrid-fmod((moved[2]-halfGrid), grid);
}

//--------------------------------------------------------------------
int	HitList::checkForHit(float x, float y, float z)
{
	int		i		= 0;
	int		numHits = 0;
	float	*tmp	= 0;
	
	if(activePrimitive)
	{
		if(activePrimitive->hasHit2)
		{
			tmp = activePrimitive->hit2;
			if(checkForHitPoint(tmp, x, y, z))
			{
				activePrimitive->wasHit = 2;
				numHits++;
			}
		}
		if(activePrimitive->hasHit1)
		{
			tmp = activePrimitive->hit1;
			if(checkForHitPoint(tmp, x, y, z))
			{
				activePrimitive->wasHit = 1;
				numHits++;
			}
		}
		tmp = activePrimitive->pos;
		if(checkForHitPoint(tmp, x, y, z))
		{
			activePrimitive->wasHit = 0;
			numHits++;
		}
	}
	if(!numHits)
	{
		for(i = 0; i < (int)primitiveList.count(); i++)
		{
			tmp = primitiveList.at(i)->pos;
			if(checkForHitPoint(tmp, x, y, z))
			{
				numHits++;
				activePrimitive = primitiveList.at(i);
				activePrimitive->wasHit = 0;
			}
		}
	}
	
	if(!numHits)
		activePrimitive = 0;
	if(activePrimitive)
	{
		QCString	primName;
		switch(activePrimitive->primType)
		{
			case HitPrimitive::Undefined:
				break;
			case HitPrimitive::Light:
				if(activePrimitive->wasHit)
					primName.sprintf("%s (%d)", ((MBALight*)(activePrimitive->primTarget))->Name(), activePrimitive->wasHit);
				else
					primName.sprintf("%s", ((MBALight*)(activePrimitive->primTarget))->Name());
				g_lightGroupWidget->setActiveLight((MBALight*)(activePrimitive->primTarget));
				emit activePrimName( primName );
				break;
			case HitPrimitive::Action:
				if(activePrimitive->wasHit)
					primName.sprintf("%s (%d)", (const char*)((FlowParticleAction*)(activePrimitive->primTarget))->Name(), activePrimitive->wasHit );
				else
					primName.sprintf("%s", (const char*)((FlowParticleAction*)(activePrimitive->primTarget))->Name() );
//				g_particleControlWidget->setActiveAction((FlowParticleAction*)(activePrimitive->primTarget));
				g_particleControlWidget->deferActiveAction((FlowParticleAction*)(activePrimitive->primTarget));
				emit activePrimName( primName );
				break;
			case HitPrimitive::Camera:
				emit activePrimName( "Camera" );
				break;
		}
	}
	else
		emit activePrimName(" ");
	return numHits;
}

//--------------------------------------------------------------------
bool	HitList::checkForHitPoint(float *point, float x, float y, float z)
{
	bool	hit = false;
	float	hB = hitBox*fov;
	if(!x)			// hit in yz plane
	{
		if(point[1] < y+hB && point[1] > y-hB)
			if(point[2] < z+hB && point[2] > z-hB)
				hit = true;
	}
	else if(!y)		// hit in xz plane
	{
		if(point[0] < x+hB && point[0] > x-hB)
			if(point[2] < z+hB && point[2] > z-hB)
				hit = true;
	}
	else if(!z)		// hit in xy plane
	{
		if(point[0] < x+hB && point[0] > x-hB)
			if(point[1] < y+hB && point[1] > y-hB)
				hit = true;
	}
	return hit;
}

//--------------------------------------------------------------------
void	HitList::drawGL()
{
	int		i;
	float	primPos[3];
	float	foreColor[4] = { 0.0, 0.0, 0.7 };
	float	backColor[4] = { 0.9, 0.9, 0.9 };
	HitPrimitive	*hP;
	
	for(i = 0; i < (int)primitiveList.count(); i++)
	{
		hP = primitiveList.at(i);
		if(hP->primType != HitPrimitive::Camera)
		{
			primPos[0] = hP->pos[0];
			primPos[1] = hP->pos[1];
			primPos[2] = hP->pos[2];
			if( hilightHitZones && g_glThick )
			{
				glLineWidth(3);
				glColor3fv(backColor);
				drawGLBounds(primPos);
			}
			glLineWidth(1);
			glColor3fv(foreColor);
			drawGLBounds(primPos);
		}
	}
	
	if(activePrimitive)
	{
		if(activePrimitive->hasHit2)
		{
			primPos[0] = activePrimitive->hit2[0];
			primPos[1] = activePrimitive->hit2[1];
			primPos[2] = activePrimitive->hit2[2];
			if( hilightHitZones && g_glThick  )
			{
				glLineWidth(3);
				glColor3fv(backColor);
				drawGLBounds(primPos);
			}
			glLineWidth(1);
			glColor3f(0.1, 1.0, 1.0);
			drawGLBounds(primPos);
		}
		if(activePrimitive->hasHit1)
		{
			primPos[0] = activePrimitive->hit1[0];
			primPos[1] = activePrimitive->hit1[1];
			primPos[2] = activePrimitive->hit1[2];
			if( hilightHitZones && g_glThick )
			{
				glLineWidth(3);
				glColor3fv(backColor);
				drawGLBounds(primPos);
			}
			glLineWidth(1);
			glColor3f(0.1, 1.0, 0.1);
			drawGLBounds(primPos);
		}
		primPos[0] = activePrimitive->pos[0];
		primPos[1] = activePrimitive->pos[1];
		primPos[2] = activePrimitive->pos[2];
		glColor3f(1.0, 1.0, 1.0);
		drawGLBounds(primPos);
	}
	
}

//--------------------------------------------------------------------
void	HitList::drawGLBounds(float *primPos)
{
	float hB = hitBox*fov;
//	warning("fovSize = %f", fovSize);

	glBegin(GL_LINE_LOOP);
	glVertex3f( primPos[0]-hB, primPos[1], primPos[2]+hB );
	glVertex3f( primPos[0]+hB, primPos[1], primPos[2]+hB );
	glVertex3f( primPos[0]+hB, primPos[1], primPos[2]-hB );
	glVertex3f( primPos[0]-hB, primPos[1], primPos[2]-hB );
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3f( primPos[0], primPos[1]-hB, primPos[2]+hB );
	glVertex3f( primPos[0], primPos[1]+hB, primPos[2]+hB );
	glVertex3f( primPos[0], primPos[1]+hB, primPos[2]-hB );
	glVertex3f( primPos[0], primPos[1]-hB, primPos[2]-hB );
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3f( primPos[0]-hB, primPos[1]+hB, primPos[2] );
	glVertex3f( primPos[0]+hB, primPos[1]+hB, primPos[2] );
	glVertex3f( primPos[0]+hB, primPos[1]-hB, primPos[2] );
	glVertex3f( primPos[0]-hB, primPos[1]-hB, primPos[2] );
	glEnd();
}












