/**
 * mbaRMShader.cpp
 * Copyright 1999 by Mark B. Allan
 * 
 * contains and activates RM shaders
 */
#include "mbaRMShader.h"

#ifndef WIN32
	#include <unistd.h>
#endif

#include "extern.h"

//==============================================================================
ShaderVariable::ShaderVariable(VariableType t)
{
	type = t;
	variable = 0;
}

ShaderVariable::ShaderVariable(ShaderVariable &in)
{
	type		= in.type;
	name		= (const char*)in.name;
	switch(type)
	{
		case Float:
			//warning("	Float:");
			variable				= new float;
			*((float*)variable)		= *((float*)in.variable);
			break;
		case Color:
			//warning("	Color:");
			variable				= new GColor;
			*((GColor*)variable)	= *((GColor*)in.variable);
			break;
		case Point:
			//warning("	Point:");
			variable				= new MVertex;
			*((MVertex*)variable)	= *((MVertex*)in.variable);
			break;
		case String:
			//warning("	String:");
			variable				= new char[strlen((char*)in.variable)+1];
			strcpy((char*)variable, (char*)in.variable);
			break;
	}
}

ShaderVariable::~ShaderVariable()
{
	switch(type)
	{
		case Float:
			delete (float*)variable;
			break;
		case Color:
			delete (GColor*)variable;
			break;
		case Point:
			delete (MVertex*)variable;
			break;
		case String:
			delete [] ((char*)variable);
			break;
	}
}

//==============================================================================
ShaderVariableDeclaration::ShaderVariableDeclaration(const char *inName, const char* inType)
{
	name = new char[strlen(inName)+1];
	type = new char[strlen(inType)+1];
	
	strcpy(name, inName);
	strcpy(type, inType);
}

ShaderVariableDeclaration::ShaderVariableDeclaration(ShaderVariableDeclaration& in)
{
	name = new char[strlen(in.name)+1];
	type = new char[strlen(in.type)+1];
	
	strcpy(name, in.name);
	strcpy(type, in.type);
}

ShaderVariableDeclaration::~ShaderVariableDeclaration()
{
	delete [] name;
	delete [] type;
}

void ShaderVariableDeclaration::declareRIB()
{
	RiDeclare (name, type );
}

//==============================================================================
MBA_RMShader::MBA_RMShader (ShaderTypeEnum inType)
{
	char	command[128];
	char	*defaultShader = 0;
	enabled			= true;
	shaderType		= Undefined;
	shadowVal		= rm_Os;
	trueDisplacement= false;
	displaceBound	= 0.1f;
	displaceCoord	= rm_object;
	color			= 0;
	opacity			= 0;
	name			= "";
	type			= "";
	visibility 		= 7;
	
	switch (inType)
	{
		case Surface:
			defaultShader = g_defaultSurfaceShader;
			break;
		case Displacement:
			defaultShader = g_defaultDisplaceShader;
			break;
		case Undefined:
			return;
	}

	FILE	*defaultFile;
	sprintf(command, "%s %s", g_shaderInfoExec, defaultShader);
//	warning("%s", command);
	#ifndef WIN32
	defaultFile = popen(command, "r");
	#else
	defaultFile = _popen(command, "r");
	#endif
	if (defaultFile > 0)
	{	
		MBA_RMShaderConstructor(defaultFile);
		#ifndef WIN32
		pclose(defaultFile);
		#else
		_pclose(defaultFile);
		#endif
	}
	else
	{
		warning("ERROR! - MBA_RMShader::MBA_RMShader() cannot open \"%s\" for read. Expect unexpected behavior...", defaultShader);
	}
}

//------------------------------------------------------------------
MBA_RMShader::MBA_RMShader (const char *shaderName)
{
	char	command[128];
	enabled			= true;
	shaderType		= Undefined;
	shadowVal		= rm_Os;
	trueDisplacement= false;
	displaceBound	= 0.1f;
	displaceCoord	= rm_object;
	color			= 0;
	opacity			= 0;
	name			= "";
	type			= "";
	visibility		= 7;
	
	FILE	*defaultFile;
	sprintf(command, "%s %s", g_shaderInfoExec, shaderName);
//	warning("%s", command);
	#ifndef WIN32
	defaultFile = popen(command, "r");
	#else
	defaultFile = _popen(command, "r");
	#endif
	if (defaultFile > 0)
	{	
		MBA_RMShaderConstructor(defaultFile);
		#ifndef WIN32
		pclose(defaultFile);
		#else
		_pclose(defaultFile);
		#endif
	}
	else
	{
		warning("ERROR! - MBA_RMShader::MBA_RMShader() cannot open \"%s\" for read. Expect unexpected behavior...", shaderName);
	}
}

//------------------------------------------------------------------
MBA_RMShader::MBA_RMShader (FILE *infile)
{
	enabled			= true;
	shaderType		= Undefined;
	shadowVal		= rm_Os;
	trueDisplacement= false;
	displaceBound	= 0.1f;
	displaceCoord	= rm_object;
	color			= 0;
	opacity			= 0;
	name			= "";
	type			= "";
	visibility 		= 7;
	
	MBA_RMShaderConstructor(infile);
}

//------------------------------------------------------------------
MBA_RMShader::MBA_RMShader (MBA_RMShader& in)
{
	enabled		= in.enabled;
	shaderType	= in.shaderType;
	shadowVal	= in.shadowVal;
	trueDisplacement = in.trueDisplacement;
	displaceBound	= in.displaceBound;
	displaceCoord	= in.displaceCoord;
	visibility		= in.visibility;
	
	declarationList.setAutoDelete(true);
	stringList.setAutoDelete(true);
	colorList.setAutoDelete(true);
	pointList.setAutoDelete(true);
	floatList.setAutoDelete(true);
	
	if (in.Color())
		color	= new GColor( *(in.Color()) );
	else
		color = 0;
		
	if (in.Opacity())
		opacity	= new GColor( *(in.Opacity()) );
	else
		opacity = 0;
		
	name		= (const char*)in.name;
	type		= (const char*)in.type;

	int i;
	for(i = 0; i < (int)in.floatList.count(); i++)
	{
		//warning("MBA_RMShader::floatList.append(%d)", i);
		floatList.append( new ShaderVariable(*(in.floatList.at(i)) ) );
	}
	for(i = 0; i < (int)in.colorList.count(); i++)
	{
		//warning("MBA_RMShader::colorList.append(%d)", i);
		colorList.append( new ShaderVariable(*(in.colorList.at(i)) ) );
	}
	for(i = 0; i < (int)in.pointList.count(); i++)
	{
		//warning("MBA_RMShader::pointList.append(%d)", i);
		pointList.append( new ShaderVariable(*(in.pointList.at(i)) ) );
	}
	for(i = 0; i < (int)in.stringList.count(); i++)
	{
		//warning("MBA_RMShader::stringList.append(%d)", i);
		stringList.append( new ShaderVariable(*(in.stringList.at(i)) ) );
	}
	for(i = 0; i < (int)in.declarationList.count(); i++)
		declarationList.append( new ShaderVariableDeclaration( *(in.declarationList.at(i)) ) );
}

//------------------------------------------------------------------
void MBA_RMShader::MBA_RMShaderConstructor (FILE *infile)
{
	char			buffer[512];
	int				i;
	QCString		tempString;
	QCString		bufferString;
	QCString		variableName;
	QCString		typeString;
	ShaderVariable	*shaderVariable;
	
	declarationList.setAutoDelete(true);
	stringList.setAutoDelete(true);
	colorList.setAutoDelete(true);
	pointList.setAutoDelete(true);
	floatList.setAutoDelete(true);
	
	// initialize variables ----------------------
	enabled				= true;
	color				= 0;
	opacity				= 0;
	type				= "";
	trueDisplacement	= false;
	displaceBound		= 0.1f;
	displaceCoord		= rm_object;
	
	while ( fgets(buffer, 512, infile) > 0)
	{
		bufferString = buffer;
		bufferString = (const char*) bufferString.simplifyWhiteSpace();
//		warning ("Reading shader:    \"%s\"", (const char*)bufferString);
		// Stop reading here...
		if (bufferString.find('}') == 0)		
			break;
		if (bufferString.find('{') == 0)		
			continue;
		if (bufferString.find("##RenderMan") == 0)		
			break;
		// if the first non-whitespace character is a '#', skip that line
		if (bufferString.find('#') == 0)		
			continue;
		if (bufferString.length() < 1)
			continue;
			
		// separate first word from bufferString and place in tempString
		tempString = (const char*)bufferString;
		i = tempString.find (' ');	//find first whitespace
		if ( i >= 0)				//isolate first word and remove it from buffer
		{
			tempString.truncate(i);
			bufferString.remove (0, i+1);
		}
		
		int		tmpInt;
		float	tmpFloat;
		// Read in color & opacity and shader type first, then get instance variables...
		if (type.isEmpty())
		{
			if (tempString == "color")			// if color info, read in and set
			{
				ReadColor(bufferString);
				continue;
			}
			else if (tempString == "opacity")	// if opacity info, read in and set
			{
				ReadOpacity(bufferString);
				continue;
			}
			else if (tempString == "visibility")	// read in shadowval
			{
				sscanf((const char*)bufferString, "[ %d ]", &tmpInt);
				visibility = tmpInt;
				continue;
			}
			else if (tempString == "shadowval")	// read in shadowval
			{
				sscanf((const char*)bufferString, "[ %d ]", &tmpInt);
				shadowVal = (ShadowTypeEnum)tmpInt;
				continue;
			}
			else if (tempString == "truedisplacement")	// read in truedisplacement
			{
				sscanf((const char*)bufferString, "[ %f ]", &tmpFloat);
				trueDisplacement = tmpFloat;
				continue;
			}
			else if (tempString == "displacebound")	// read in truedisplacement
			{
				sscanf((const char*)bufferString, "[ %f ]", &tmpFloat);
				displaceBound = tmpFloat;
				continue;
			}
			else if (tempString == "displacecoord")	// read in truedisplacement
			{
				sscanf((const char*)bufferString, "[ %d ]", &tmpInt);
				displaceCoord = (CoordSystemEnum)tmpInt;
				continue;
			}
			else if (tempString == "surface")	// assign type and get shader name
			{
				//if (shaderType != Surface)
				//	warning ("ERROR READING SHADER: Requested as %d, but assigned to Surface (%d)", shaderType, Surface);
				shaderType = Surface;
				ReadIsSurface(bufferString);
				continue;
			}
			else if (tempString == "displacement")	// assign type and get shader name
			{
				//if (shaderType != Displacement)
				//	warning ("ERROR READING SHADER: Requested as %d, but assigned to Displacement (%d)", shaderType, Displacement);
				shaderType = Displacement;
				ReadIsDisplacement(bufferString);
				continue;
			}
			else
			{
				warning ("unrecognized token %s --- %s ", (const char*)tempString, (const char*)bufferString);
			}
		}
		else
		{
		// get shader instance variables ------------------------
			variableName	= (const char*)NoQuotes(tempString);
			typeString		= (const char*)NoQuotes(bufferString);
			
			fgets(buffer, 512, infile);
			
			bufferString = buffer;
			bufferString = bufferString.simplifyWhiteSpace();
			i = bufferString.findRev(":");					//remove everything up to and including ':'
			bufferString.remove(0, i+1);
			bufferString = bufferString.simplifyWhiteSpace();
			
			if (typeString.find("float") >= 0)
			{
				shaderVariable = new ShaderVariable(ShaderVariable::Float);
				shaderVariable->name = (const char*)variableName;
				shaderVariable->variable = new float;
				{	
					// read in value
					*((float*)shaderVariable->variable) = bufferString.toFloat();
				}
				floatList.append(shaderVariable);
				declarationList.append(new 	ShaderVariableDeclaration(variableName, "uniform float"));
			}
			else if (typeString.find("color") >= 0)
			{
				shaderVariable = new ShaderVariable(ShaderVariable::Color);
				shaderVariable->name = (const char*)variableName;
				shaderVariable->variable = new GColor ();
				{
					// read in value
					RtColor	rtColor;
					GetPoint ((const char*)bufferString, rtColor);
					((GColor*)shaderVariable->variable)->set(rtColor);
				}
				colorList.append(shaderVariable);
				declarationList.append(new 	ShaderVariableDeclaration(variableName, "uniform color"));
			}
			else if (typeString.find("point") >= 0)
			{
				shaderVariable = new ShaderVariable(ShaderVariable::Point);
				shaderVariable->name = (const char*)variableName;
				shaderVariable->variable = new MVertex ();
				{		
					// read in value
					RtPoint	rtPoint;
					GetPoint ((const char*)bufferString, rtPoint);
					((MVertex*)shaderVariable->variable)->set(rtPoint);
				}
				pointList.append(shaderVariable);
				declarationList.append(new 	ShaderVariableDeclaration(variableName, "uniform point"));
			}
			else if (typeString.find("string") >= 0)
			{
				shaderVariable = new ShaderVariable(ShaderVariable::String);
				shaderVariable->name = (const char*)variableName;
				//warning("sending (%s) to NoQuotes", (const char*)bufferString);
				shaderVariable->variable = new char[strlen((const char*)NoQuotes(bufferString))+1];
				strcpy((char*)(shaderVariable->variable), (const char*)NoQuotes(bufferString));
				stringList.append(shaderVariable);
				declarationList.append(new 	ShaderVariableDeclaration(variableName, "uniform string"));
			}
		}
		if (strrchr(buffer, EOF))
			break;
	}
}

//------------------------------------------------------------------
MBA_RMShader& MBA_RMShader::operator =( MBA_RMShader& in )
{
	shaderType = in.shaderType;
	trueDisplacement = in.trueDisplacement;
	
	visibility	= in.visibility;
	
	if (color)
		delete color;
	if (opacity)
		delete opacity;
		
	if (in.Color())
		color	= new GColor( *(in.Color()) );
	else
		color = 0;
		
	if (in.Opacity())
		opacity	= new GColor( *(in.Opacity()) );
	else
		opacity = 0;
		
	name		= (const char*)in.name;
	type		= (const char*)in.type;
	floatList.clear();
	colorList.clear();
	pointList.clear();
	stringList.clear();
	declarationList.clear();
	int i;
	for(i = 0; i < (int)in.floatList.count(); i++)
	{
		//warning("MBA_RMShader::floatList.append(%d)", i);
		floatList.append( new ShaderVariable(*(in.floatList.at(i)) ) );
	}
	for(i = 0; i < (int)in.colorList.count(); i++)
	{
		//warning("MBA_RMShader::colorList.append(%d)", i);
		colorList.append( new ShaderVariable(*(in.colorList.at(i)) ) );
	}
	for(i = 0; i < (int)in.pointList.count(); i++)
	{
		//warning("MBA_RMShader::pointList.append(%d)", i);
		pointList.append( new ShaderVariable(*(in.pointList.at(i)) ) );
	}
	for(i = 0; i < (int)in.stringList.count(); i++)
	{
		//warning("MBA_RMShader::stringList.append(%d)", i);
		stringList.append( new ShaderVariable(*(in.stringList.at(i)) ) );
	}
	for(i = 0; i < (int)in.declarationList.count(); i++)
		declarationList.append( new ShaderVariableDeclaration( *(in.declarationList.at(i)) ) );
	return *this;
}

//------------------------------------------------------------------
MBA_RMShader::~MBA_RMShader()
{
	if (color)
		delete color;
	if (opacity)
		delete opacity;
}

//-------------------------------------------------------------------
GColor& MBA_RMShader::colorsAt (int i)		
{
	static	GColor	dummy;
	if (i < numColors())
		return	*((GColor*)colorList.at(i)->variable);	
	else
	{
		warning("ERROR: Out of range asking for MBA_RMShader::colorsAt(%d), numColors = %d", i, numColors());
		warning("	Invalid GColor reference returned.");
		return dummy;
	}
}
	
//-------------------------------------------------------------------
float& MBA_RMShader::floatsAt (int i)		
{
	static	float	dummy = 0.0;
	if (i < numFloats())
		return	*((float*)floatList.at(i)->variable);	
	else
	{
		warning("ERROR: Out of range asking for MBA_RMShader::floatsAt(%d), numFloats = %d", i, numFloats());
		warning("	Invalid float reference returned.");
		return dummy;
	}
}
	
//-------------------------------------------------------------------
MVertex& MBA_RMShader::pointsAt (int i)		
{
	static	MVertex	dummy;
	if (i < numPoints())
		return	*((MVertex*)pointList.at(i)->variable);	
	else
	{
		warning("ERROR: Out of range asking for MBA_RMShader::pointsAt(%d), numPoints = %d", i, numPoints());
		warning("	Invalid MVertex reference returned.");
		return dummy;
	}
}
	
//-------------------------------------------------------------------
char* MBA_RMShader::stringsAt (int i)		
{
	if (i < numStrings())
		return	(char*)(stringList.at(i)->variable);
	else
	{
		warning("ERROR: Out of range asking for MBA_RMShader::stringsAt(%d), numStrings = %d", i, numStrings());
		warning("	Invalid QCString reference returned.");
		return "";
	}
}
	
//-------------------------------------------------------------------
void MBA_RMShader::DeclareVariables()
{
	int i;
	for (i = 0; i < (int)declarationList.count(); i++)
		declarationList.at(i)->declareRIB();
}

//-------------------------------------------------------------------
void	MBA_RMShader::SetFloat (int index, float val)
{
	if (index < numFloats())
	{
		*((float*)floatList.at(index)->variable) = val;
	}
}

//-------------------------------------------------------------------
void	MBA_RMShader::setColor (GColor &in)
{
	if (color)
		*color = in;
}

//-------------------------------------------------------------------
void	MBA_RMShader::SetOpacity (GColor &in)
{
	if (opacity)
		*opacity = in;
}

//-------------------------------------------------------------------
void	MBA_RMShader::setColor (int index, GColor &in)
{	
	if (index < numColors())
		*((GColor*)colorList.at(index)->variable) = in;		
}

//-------------------------------------------------------------------
void	MBA_RMShader::SetPoint (int index, MVertex &in)
{	
	if (index < numPoints())
		*((MVertex*)pointList.at(index)->variable) = in;	
}

//-------------------------------------------------------------------
void	MBA_RMShader::SetPoint0(int index, float in)
{	
	if (index < numPoints())
		(*((MVertex*)pointList.at(index)->variable)).setX(in);
}

//-------------------------------------------------------------------
void	MBA_RMShader::SetPoint1(int index, float in)
{	
	if (index < numPoints())
		(*((MVertex*)pointList.at(index)->variable)).setY(in);
}

//-------------------------------------------------------------------
void	MBA_RMShader::SetPoint2(int index, float in)
{
	if (index < numPoints())
		(*((MVertex*)pointList.at(index)->variable)).setZ(in);
}

//-------------------------------------------------------------------
void	MBA_RMShader::SetString(int index, const char* val)
{
	if (index < numStrings())
	{
		delete [] (char*)stringList.at(index)->variable;
		stringList.at(index)->variable = new char[strlen(val)+1];
		strcpy((char*)(stringList.at(index)->variable), val);
	}
}

//-------------------------------------------------------------------
void MBA_RMShader::ReadColor (QCString& string)
{
	float	rtColor[3];
	
	if (!color)
		color	= new GColor(1.0, 1.0, 1.0);
	GetPoint ((const char*)string, rtColor);
	color->SetRGB(rtColor[0], rtColor[1], rtColor[2]);
//	warning ("set color : %f %f %f", rtColor[0], rtColor[1], rtColor[2]);
//	warning (" MBA_RMShader::ReadColor(%s)", (const char*)string);
}

//-------------------------------------------------------------------
void MBA_RMShader::ReadOpacity (QCString& string)
{
	float	rtColor[3];
	
	if ( !opacity)
		opacity	= new GColor(1.0, 1.0, 1.0);
	GetPoint ((const char*)string, rtColor);
	opacity->set(rtColor);
//	warning (" MBA_RMShader::ReadOpacity(%s)", (const char*)string);
}

//-------------------------------------------------------------------
void MBA_RMShader::ReadIsSurface (QCString& string)
{
	type = "surface";
	if (!color)
		color	= new GColor(1.0, 1.0, 1.0);
	if (!opacity)
		opacity	= new GColor(1.0, 1.0, 1.0);
	name = (const char*)NameOnly(string);
//	warning (" MBA_RMShader::ReadIsSurface(%s)", (const char*)string);
}

//-------------------------------------------------------------------
void MBA_RMShader::ReadIsDisplacement (QCString& string)
{
	type = "displacement";
	name = (const char*)NameOnly(string);
//	warning (" MBA_RMShader::ReadIsDisplacement(%s)", (const char*)string);
}
	
//-------------------------------------------------------------------
QCString MBA_RMShader::NoQuotes (QCString& string)
{	
	int i;
	QCString	tmpString = (const char*) string;
	//warning("--No quotes");
	i = tmpString.find("\"");
	//warning("i = %d", i);
	if (i >= 0)	
		tmpString.remove(0, i+1);
	else
		warning ("MBA_RMShader::NoQuotes error: no opening quote (%s)", (const char*)string);
	i = tmpString.findRev("\"");
	//warning("i = %d", i);
	if (i >= 0)
		tmpString.truncate(i);
	else
		warning ("MBA_RMShader::NoQuotes error: no closing quote (%s)", (const char*)string);

	return tmpString;
}

//-------------------------------------------------------------------
QCString MBA_RMShader::NameOnly (QCString& string)
{	
	int i;
	QCString	tmpString = (const char*)string;
	
	i = tmpString.findRev("/");
	if(i >= 0)
		tmpString.remove(0, i+1);
		
	i = tmpString.find(".");
	if(i >= 0)
		tmpString.truncate(i);
	
	i = tmpString.find("\"");
	if(i >= 0)	
		tmpString.remove(0, i+1);

	i = tmpString.findRev("\"");
	if(i >= 0)
		tmpString.truncate(i);

	return tmpString;
}

//-------------------------------------------------------------------
void MBA_RMShader::GetPoint (const char* str, float point[3])
{
	int		i;
	QCString	buffer(str);
	QCString	r;
	QCString	g;
	QCString	b;
	
	point[0] = -1.0;	point[1] = -1.0;	point[2] = -1.0;
	
	i = buffer.find('[');
	if ( i < 0)	{	warning("([) Error in passing string \"%s\" to Shader::GetPoint", str); return; }
	buffer.remove (0, i+1);
	
	i = buffer.find(']');
	if ( i < 0)	{	warning("(]) Error in passing string \"%s\" to Shader::GetPoint", str); return; }
	buffer.truncate (i);
	
	buffer = buffer.simplifyWhiteSpace();
//	warning("MBA_RMShader::GetPoint - buffer = \"%s\"", (const char*)buffer);
	
	r = (const char*)buffer;
	i = buffer.find(' ');
	if ( i < 0)	{	warning("(r) Error in passing string \"%s\" to Shader::GetPoint", (const char*)buffer); return; }
	r.truncate(i);
	buffer.remove(0, i+1);
//	warning("MBA_RMShader::GetPoint - buffer = \"%s\"", (const char*)buffer);
	
	g = (const char*)buffer;
	i = buffer.find(' ');
	if ( i < 0)	{	warning("(g) Error in passing string \"%s\" to Shader::GetPoint", (const char*)buffer); return; }
	g.truncate(i);
	buffer.remove(0, i+1);
//	warning("MBA_RMShader::GetPoint - buffer = \"%s\"", (const char*)buffer);

	b = (const char*)buffer;

//	warning(" %s %s %s ", (const char*)r, (const char*)g, (const char*)b );
	point[0] = r.toFloat();
	point[1] = g.toFloat();
	point[2] = b.toFloat();	
}

//-------------------------------------------------------------------
void MBA_RMShader::Activate ()
{
	char	buffer[512];
	char	*variable[100];
	void	*value[100];
	int		i;
	int		vcount;
	RtColor	rcolor;
	
	if (!enabled)
		return;
		
	char	*coordCamera  = "camera";
	char	*coordCurrent = "current";
	char	*coordNull    = "null";
	char	*coordObject  = "object";
	char	*coordScreen  = "screen";
	char	*coordShader  = "shader";
	char	*coordWorld   = "world";

	char	*shadowNone   = "none";
	char	*shadowOpaque = "opaque";
	char	*shadowOs	  = "Os";
	char	*shadowShade  = "shade";

	RtFloat	dBound = displaceBound;
	switch(shaderType)
	{
		case Surface:
			if (color)
			{
				rcolor[0] = color->R();
				rcolor[1] = color->G();
				rcolor[2] = color->B();
				RiColor (rcolor);
			}
			if (opacity)
			{
				rcolor[0] = opacity->R();
				rcolor[1] = opacity->G();
				rcolor[2] = opacity->B();
				RiOpacity (rcolor);
			}
			// FIXME - 
			// FIXME - if( we're using BMRT... )
			// FIXME - {
				switch (shadowVal)
				{
					case rm_none:
						RiAttribute( "render", "casts_shadows", &shadowNone, RI_NULL );
						break;
					case rm_opaque:
						RiAttribute( "render", "casts_shadows", &shadowOpaque, RI_NULL );
						break;
					case rm_Os:
						RiAttribute( "render", "casts_shadows", &shadowOs, RI_NULL );
						break;
					case rm_shade:
						RiAttribute( "render", "casts_shadows", &shadowShade, RI_NULL );
						break;
				}
				RiAttribute("render", "visibility", &visibility, RI_NULL);
				//warning("render visibility = %d", visibility);
			// FIXME - }
			break;
		case Displacement:
			if ( trueDisplacement && g_allowTrueDisplace )
			{
				RiAttribute ("render", "truedisplacement", &trueDisplacement, RI_NULL);
				switch (displaceCoord)
				{
					case rm_camera:
						RiAttribute("displacementbound", "coordinatesystem", (RtPointer)&coordCamera, "sphere", (RtPointer)&dBound, RI_NULL);					
						break;
					case rm_current:
						RiAttribute("displacementbound", "coordinatesystem", (RtPointer)&coordCurrent, "sphere", (RtPointer)&dBound, RI_NULL);					
						break;
					case rm_null:
						RiAttribute("displacementbound", "coordinatesystem", (RtPointer)&coordNull, "sphere", (RtPointer)&dBound, RI_NULL);					
						break;
					case rm_object:
						RiAttribute("displacementbound", "coordinatesystem", (RtPointer)&coordObject, "sphere", (RtPointer)&dBound, RI_NULL);					
						break;
					case rm_screen:
						RiAttribute("displacementbound", "coordinatesystem", (RtPointer)&coordScreen, "sphere", (RtPointer)&dBound, RI_NULL);					
						break;
					case rm_shader:
						RiAttribute("displacementbound", "coordinatesystem", (RtPointer)&coordShader, "sphere", (RtPointer)&dBound, RI_NULL);					
						break;
					case rm_world:
						RiAttribute("displacementbound", "coordinatesystem", (RtPointer)&coordWorld, "sphere", (RtPointer)&dBound, RI_NULL);					
						break;
				}
			}
			break;
		case Undefined:
			warning ("ERROR! Trying to Activate() and Undefined shader!");
			return;
	}

	//-- Initialize instance variable list -----------------
	for (i = 0; i < 100; i++)
	{
		value[i]	= RI_NULL;
		variable[i]	= (RtToken)RI_NULL;
	}
	//--
	vcount = 0;
	for (i = 0; i < numFloats(); i++)
	{
		variable[vcount]	= new char[strlen(floatsNameAt(i))+1];
		strcpy ( variable[vcount], floatsNameAt(i) );
		value[vcount]		= floatsValAt(i);
		vcount++;
	}		
	for (i = 0; i < numColors(); i++)
	{
		variable[vcount]	= new char[strlen(colorsNameAt(i))+1];
		strcpy ( variable[vcount], colorsNameAt(i) );
		value[vcount]		= colorsValAt(i);
		vcount++;
	}
	for (i = 0; i < numPoints(); i++)
	{
		variable[vcount]	= new char[strlen(pointsNameAt(i))+1];
		strcpy ( variable[vcount], pointsNameAt(i) );
		value[vcount]		= pointsValAt(i);
		vcount++;
	}
	for (i = 0; i < numStrings(); i++)
	{
		//warning("string %d = (%s)", i, stringsValAt(i));
		variable[vcount]	= new char[strlen(stringsNameAt(i))+1];
		strcpy ( variable[vcount], stringsNameAt(i) );
		value[vcount]		= stringsValAt(i);
		vcount++;
	}
	
	//-- This is a hack, but how to do it properly? --------
	strcpy (buffer, name);
	if (type == "surface")
		RiSurface (buffer,
		variable[0],  value[0], 
		variable[1],  value[1],
		variable[2],  value[2],
		variable[3],  value[3],
		variable[4],  value[4],
		variable[5],  value[5],
		variable[6],  value[6],
		variable[7],  value[7],
		variable[8],  value[8],
		variable[9],  value[9],
		variable[10], value[10],
		variable[11], value[11],
		variable[12], value[12],
		variable[13], value[13],
		variable[14], value[14],
		variable[15], value[15],
		variable[16], value[16],
		variable[17], value[17],
		variable[18], value[18],
		variable[19], value[19],
		variable[20], value[20],
		variable[21], value[21],
		variable[22], value[22],
		variable[23], value[23],
		variable[24], value[24],
		variable[25], value[25],
		variable[26], value[26],
		variable[27], value[27],
		variable[28], value[28],
		variable[29], value[29],
		variable[30], value[30],
		variable[31], value[31],
		variable[32], value[32],
		variable[33], value[33],
		variable[34], value[34],
		variable[35], value[35],
		variable[36], value[36],
		variable[37], value[37],
		variable[38], value[38],
		variable[39], value[39],
		variable[40], value[40],
		variable[41], value[41],
		variable[42], value[42],
		variable[43], value[43],
		variable[44], value[44],
		variable[45], value[45],
		variable[46], value[46],
		variable[47], value[47],
		variable[48], value[48],
		variable[49], value[49],
		variable[50], value[50],
		variable[51], value[51],
		variable[52], value[52],
		variable[53], value[53],
		variable[54], value[54],
		variable[55], value[55],
		variable[56], value[56],
		variable[57], value[57],
		variable[58], value[58],
		variable[59], value[59],
		variable[60], value[60],
		variable[61], value[61],
		variable[62], value[62],
		variable[63], value[63],
		variable[64], value[64],
		variable[65], value[65],
		variable[66], value[66],
		variable[67], value[67],
		variable[68], value[68],
		variable[69], value[69],
		variable[70], value[70],
		variable[71], value[71],
		variable[72], value[72],
		variable[73], value[73],
		variable[74], value[74],
		variable[75], value[75],
		variable[76], value[76],
		variable[77], value[77],
		variable[78], value[78],
		variable[79], value[79],
		variable[80], value[80],
		variable[81], value[81],
		variable[82], value[82],
		variable[83], value[83],
		variable[84], value[84],
		variable[85], value[85],
		variable[86], value[86],
		variable[87], value[87],
		variable[88], value[88],
		variable[89], value[89],
		variable[90], value[90],
		variable[91], value[91],
		variable[92], value[92],
		variable[93], value[93],
		variable[94], value[94],
		variable[95], value[95],
		variable[96], value[96],
		variable[97], value[97],
		variable[98], value[98],
		variable[99], value[99],
		RI_NULL);
	else if (type == "displacement")
		RiDisplacement (buffer,
		variable[0],  value[0], 
		variable[1],  value[1],
		variable[2],  value[2],
		variable[3],  value[3],
		variable[4],  value[4],
		variable[5],  value[5],
		variable[6],  value[6],
		variable[7],  value[7],
		variable[8],  value[8],
		variable[9],  value[9],
		variable[10], value[10],
		variable[11], value[11],
		variable[12], value[12],
		variable[13], value[13],
		variable[14], value[14],
		variable[15], value[15],
		variable[16], value[16],
		variable[17], value[17],
		variable[18], value[18],
		variable[19], value[19],
		variable[20], value[20],
		variable[21], value[21],
		variable[22], value[22],
		variable[23], value[23],
		variable[24], value[24],
		variable[25], value[25],
		variable[26], value[26],
		variable[27], value[27],
		variable[28], value[28],
		variable[29], value[29],
		variable[30], value[30],
		variable[31], value[31],
		variable[32], value[32],
		variable[33], value[33],
		variable[34], value[34],
		variable[35], value[35],
		variable[36], value[36],
		variable[37], value[37],
		variable[38], value[38],
		variable[39], value[39],
		variable[40], value[40],
		variable[41], value[41],
		variable[42], value[42],
		variable[43], value[43],
		variable[44], value[44],
		variable[45], value[45],
		variable[46], value[46],
		variable[47], value[47],
		variable[48], value[48],
		variable[49], value[49],
		variable[50], value[50],
		variable[51], value[51],
		variable[52], value[52],
		variable[53], value[53],
		variable[54], value[54],
		variable[55], value[55],
		variable[56], value[56],
		variable[57], value[57],
		variable[58], value[58],
		variable[59], value[59],
		variable[60], value[60],
		variable[61], value[61],
		variable[62], value[62],
		variable[63], value[63],
		variable[64], value[64],
		variable[65], value[65],
		variable[66], value[66],
		variable[67], value[67],
		variable[68], value[68],
		variable[69], value[69],
		variable[70], value[70],
		variable[71], value[71],
		variable[72], value[72],
		variable[73], value[73],
		variable[74], value[74],
		variable[75], value[75],
		variable[76], value[76],
		variable[77], value[77],
		variable[78], value[78],
		variable[79], value[79],
		variable[80], value[80],
		variable[81], value[81],
		variable[82], value[82],
		variable[83], value[83],
		variable[84], value[84],
		variable[85], value[85],
		variable[86], value[86],
		variable[87], value[87],
		variable[88], value[88],
		variable[89], value[89],
		variable[90], value[90],
		variable[91], value[91],
		variable[92], value[92],
		variable[93], value[93],
		variable[94], value[94],
		variable[95], value[95],
		variable[96], value[96],
		variable[97], value[97],
		variable[98], value[98],
		variable[99], value[99],
		RI_NULL);
		
		for (i = 0; i < vcount; i++)
		{
			delete [] variable[i];
		}
}


//-------------------------------------------------------------------
void MBA_RMShader::SaveToFile(FILE *outfile, int tabs)
{
	int		i;
	char	buffer[128];
	
	char	*walker = buffer;
	for (i = 0; i < tabs; i++)
	{	*walker = '\t';	*(walker+1) = '\0';	walker++;	}
	
	switch(shaderType)
	{
		case Surface:
			sprintf(walker, "RM_SURFACE\n");
			break;
		case Displacement:
			sprintf(walker, "RM_DISPLACEMENT\n");
			break;
		case Undefined:
			return;
	}
	fputs(buffer, outfile);
	sprintf(walker, "{\n");
	fputs(buffer, outfile);
	
	SaveInfo(outfile, tabs+1);

	sprintf(walker, "}\n");
	fputs(buffer, outfile);
}

//-------------------------------------------------------------------
void MBA_RMShader::SaveInfo(FILE *outfile, int tabs)
{
	int		i;
	char	tmpString[128];
	char	buffer[128];
	
	char	*walker = buffer;
	for (i = 0; i < tabs; i++)
	{	*walker = '\t';	*(walker+1) = '\0';	walker++;	}
	
	if ( color )
	{
		sprintf(walker, "%-16s [ %f %f %f ]\n", "color", color->R(), color->G(), color->B() );
		fputs(buffer, outfile);
	}
	if ( opacity )
	{
		sprintf(walker, "%-16s [ %f %f %f ]\n", "opacity", opacity->R(), opacity->G(), opacity->B() );
		fputs(buffer, outfile);
	}

	switch(shaderType)
	{
		case Surface:
			sprintf( walker, "%-16s [ %d ]\n", "visibility", (int)visibility );
			fputs(buffer, outfile);
			sprintf( walker, "%-16s [ %d ]\n", "shadowval", (int)shadowVal );
			fputs(buffer, outfile);
			sprintf( walker, "%-16s \"%s\"\n", "surface", Name() );
			fputs(buffer, outfile);
			break;
		case Displacement:
			sprintf( walker, "%-16s [ %f ]\n", "truedisplacement", trueDisplacement );
			fputs(buffer, outfile);
			sprintf( walker, "%-16s [ %f ]\n", "displacebound", displaceBound );
			fputs(buffer, outfile);
			sprintf( walker, "%-16s [ %d ]\n", "displacecoord", displaceCoord );
			fputs(buffer, outfile);
			sprintf (walker, "%-16s \"%s\"\n", "displacement", Name() );
			fputs(buffer, outfile);
			break;
		case Undefined:
			return;
	}
	
	//-----write floats
	for (i = 0; i < numFloats(); i++)
	{
		sprintf(tmpString, "\"%s\"", floatsNameAt(i));
		sprintf( walker, "\t%-31s \"uniform float\"\n", tmpString);
		fputs(buffer, outfile);
//		sprintf( walker, "\t\t\tDefault Value: %f\n", floatsAt(i) );
		sprintf( walker, "\t\tValue: %f\n", floatsAt(i) );
		fputs(buffer, outfile);
	}
	//-----write colors
	for (i = 0; i < numColors(); i++)
	{
		sprintf(tmpString, "\"%s\"", colorsNameAt(i));
		sprintf( walker, "\t%-31s \"uniform color\"\n", tmpString);
		fputs(buffer, outfile);
//		sprintf( walker, "\t\t\tDefault Value: \"rgb\" [ %f %f %f ]\n",  colorsAt(i).R(), colorsAt(i).G(), colorsAt(i).B() );
		sprintf( walker, "\t\tValue: \"rgb\" [ %f %f %f ]\n",  colorsAt(i).R(), colorsAt(i).G(), colorsAt(i).B() );
		fputs(buffer, outfile);                                           
	}
	//-----write point
	for (i = 0; i < numPoints(); i++)
	{
		sprintf(tmpString, "\"%s\"", pointsNameAt(i) );
		sprintf( walker, "\t%-31s \"uniform point\"\n", tmpString);
		fputs(buffer, outfile);
//		sprintf( walker, "\t\t\tDefault Value: [ %f %f %f ]\n",  pointsAt(i).X(), pointsAt(i).Y(), pointsAt(i).Z() );
		sprintf( walker, "\t\tValue: [ %f %f %f ]\n",  pointsAt(i).X(), pointsAt(i).Y(), pointsAt(i).Z() );
		fputs(buffer, outfile);
	}
	//-----write strings
	for (i = 0; i < numStrings(); i++)
	{
		sprintf(tmpString, "\"%s\"", stringsNameAt(i) );
		sprintf( walker, "\t%-31s \"uniform string\"\n", tmpString);
		fputs(buffer, outfile);
//		sprintf( walker, "\t\t\tDefault Value: \"%s\"\n",  (const char*)stringsAt(i) );
		sprintf( walker, "\t\tValue: \"%s\"\n",  (const char*)stringsAt(i) );
		fputs(buffer, outfile);
	}
}	

