/**
 * mbaRMShader.h
 * Copyright 1999 by Mark B. Allan
 * 
 * contains and activates RM shaders
 */
#ifndef mbaRMShader_h
#define mbaRMShader_h

#include <stdio.h>

#include <qobject.h>
#include <qlist.h>

#include "basetypes.h"
#include "mmath.h"

#include <ri.h>

/**
 * ShaderVariable
 */
//======================================================================
class	ShaderVariable 
{
public:
	enum VariableType { Float, Color, Point, String };
	
	ShaderVariable (VariableType t);
	ShaderVariable (ShaderVariable& in);
	~ShaderVariable ();
	
	VariableType	type;
	QCString		name;
	void			*variable;
};

/**
 * ShaderVariableDeclaration
 */
//====================================================================
class ShaderVariableDeclaration
{
public:
	ShaderVariableDeclaration(const char *inName, const char* inType);
	ShaderVariableDeclaration(ShaderVariableDeclaration& in);
	~ShaderVariableDeclaration();
	void	declareRIB();
protected:
	char*	name;
	char*	type;
};

/**
 * shader wrapper
 */
//======================================================================
class	MBA_RMShader /*: public QObject*/
{
/*    Q_OBJECT*/
	
public:
	enum ShaderTypeEnum {	Undefined = -1, Surface, Displacement };
	enum ShadowTypeEnum {	rm_none = 0, rm_opaque, rm_Os, rm_shade };
	enum CoordSystemEnum{	rm_camera = 0, rm_current, rm_null, rm_object, rm_screen, rm_shader, rm_world };
	
	MBA_RMShader (ShaderTypeEnum = Undefined);
	MBA_RMShader (const char*);
	MBA_RMShader (FILE *);
	MBA_RMShader (MBA_RMShader& in);
	~MBA_RMShader ();
	MBA_RMShader&	operator = (MBA_RMShader& in);
	
	void	MBA_RMShaderConstructor(FILE *);
	
	void	Activate ();
	void	GetPoint (const char*, float[3] );
//	void	SetParentObject (Object *);
	
	void	DeclareVariables();
	
	void	SaveToFile(FILE *outfile, int tabs);
	void	SaveInfo(FILE *outfile, int tabs = 0);
	
	inline	const	char	*Name ()	{	return (const char*)name;	}
	inline	const	char	*Type ()	{	return (const char*)type;	}
	inline	ShaderTypeEnum	ShaderType(){	return shaderType;	}
	
	inline	GColor	*Color()		{	return color;	}
	inline	GColor	*Opacity()		{	return opacity;	}
	inline	int		numFloats()		{	return floatList.count();	}
	inline	int		numColors()		{	return colorList.count();	}
	inline	int		numPoints()		{	return pointList.count();	}
	inline	int		numStrings()	{	return stringList.count();	}

	float	&floatsAt (int i);
	GColor	&colorsAt (int i);
	MVertex	&pointsAt (int i);
	char	*stringsAt(int i);
	
	inline	const	char	*floatsNameAt (int i)	{	if (i<numFloats()) return (const char*)floatList.at(i)->name;	else return 0;	}
	inline	const	char	*colorsNameAt (int i)	{	if (i<numColors()) return (const char*)colorList.at(i)->name;	else return 0;	}
	inline	const	char	*pointsNameAt (int i)	{	if (i<numPoints()) return (const char*)pointList.at(i)->name;	else return 0;	}
	inline	const	char	*stringsNameAt(int i)	{	if (i<numStrings()) return (const char*)stringList.at(i)->name;	else return 0;	}
	
	inline	float	*floatsValAt (int i)	{	return ( (float*)floatList.at(i)->variable);			}
	inline	float	*colorsValAt (int i)	{	return ((GColor*)colorList.at(i)->variable)->RibColor();}
	inline	float	*pointsValAt (int i)	{	return ((MVertex*)pointList.at(i)->variable)->XYZ();	}
	inline	char 	**stringsValAt (int i)	{	return (char**)(&(stringList.at(i)->variable));	}

	inline	void	SetShadowVal(ShadowTypeEnum	in)		{	shadowVal = in;					}
	inline	int		ShadowVal()							{	return (int)shadowVal;			}
	inline	void	SetTrueDisplacement(bool in)		{	if(in) trueDisplacement = 1.0; else trueDisplacement = 0.0;	}
	inline	bool	TrueDisplacement()					{	return (bool)trueDisplacement;	}
	inline	void	SetDisplaceBound(float in)			{	displaceBound = in;				}
	inline	float	DisplaceBound()						{	return displaceBound;			}
	inline	void	SetDisplaceCoord(CoordSystemEnum in){	displaceCoord = in;				}
	inline	int		DisplaceCoord()						{	return (int)displaceCoord;		}

	void	setColor (GColor &in);
	void	SetOpacity(GColor &in);
	void	SetFloat (int index, float val);
	void	setColor (int index, GColor &in);
	void	SetPoint (int index, MVertex &in);
	void	SetPoint0(int index, float in);
	void	SetPoint1(int index, float in);
	void	SetPoint2(int index, float in);
	void	SetString(int index, const char* val);

	inline	void	setEnabled(bool status)	{	enabled = status;	}
	inline	bool	Enabled()				{	return enabled;		}
	
	inline	bool VisCamera() 	{	return (bool)(visibility & 1);	}
	inline	bool VisReflect()	{	return (bool)(visibility & 2);	}
	inline	bool VisShadow()	{	return (bool)(visibility & 4);	}
	inline	int	 Visibility()	{	return visibility;	}
	inline	void SetVisibility(int vis)	{	warning("SetVisibility(%d)", vis); visibility = vis; }
	
//public slots:
//	void		SetTrueDisplacement(bool);

protected:
	QCString	NoQuotes(QCString&);
	QCString	NameOnly(QCString&);
	void		ReadColor	(QCString&);
	void		ReadOpacity	(QCString&);
	void		ReadIsSurface(QCString&);
	void		ReadIsDisplacement(QCString&);
	

protected:
	ShaderTypeEnum			shaderType;
	ShadowTypeEnum			shadowVal;
	float					trueDisplacement;
	float					displaceBound;
	CoordSystemEnum			displaceCoord;
	GColor					*color;
	GColor					*opacity;
	QCString				name;
	QCString				type;
	QList<ShaderVariable>	floatList;
	QList<ShaderVariable>	colorList;
	QList<ShaderVariable>	pointList;
	QList<ShaderVariable>	stringList;
	QList<ShaderVariableDeclaration>	declarationList;
	
	bool	enabled;
	int		visibility;
};

#endif

