/**
 * mbaScene.cpp
 * Copyright 1999 by Mark B. Allan
 * 
 * contains and controls the scene
 */
#include "mbaScene.h"

#include <qfile.h>
#include <qtextstream.h>
#include <qprogressdialog.h>

#include "extern.h"
#include "define.h"
#include "lightGroup.h"
#include "particleControl.h"
#include "glPortCam.h"
#include "camera.h"

//====================================================================
MBAScene::MBAScene()
{
//	backgroundColor	= new GColor(0.25, 0.25, 0.25);
	backgroundColor	= new GColor(0.0, 0.0, 0.0);
	
	cameraList		= new QList<CameraView>;
	lightGroupList	= new QList<LightGroup>;
	particleList	= new QList<ParticleControl>;
	
	glCullFace		= false;
	glBlend			= false;
	glTwoSideLight	= false;
	
	activeLightGroup		= -1;
	activeCamera			= -1;
	activeParticleSystem	= -1;
	//??? Should scene autodelete objects? I think so.
	particleList->setAutoDelete(true);
	lightGroupList->setAutoDelete(true);
}

MBAScene::~MBAScene()
{
	//warning("MBAScene::~MBAScene()");
	cameraList->clear();
	lightGroupList->clear();
	particleList->clear();
	delete cameraList;
	delete lightGroupList;
	delete particleList;
}

//--------------------------------------------------------------------
void	MBAScene::clear()
{
	cameraList->clear();
	lightGroupList->clear();
	particleList->clear();
	activeLightGroup		= -1;
	activeCamera			= -1;
	activeParticleSystem	= -1;
}

//--------------------------------------------------------------------
void	MBAScene::reset()
{
	int i;
	for (i = 0; i < (int)particleList->count(); i++)
		particleList->at(i)->ResetSystem();
	g_frame = 0;
}

//--------------------------------------------------------------------
void	MBAScene::execute(int frame)
{
	for (int i = 0; i < (int)particleList->count(); i++)
		particleList->at(i)->Execute(frame);
}

//--------------------------------------------------------------------
void	MBAScene::drawGL(int frame )
{
	int i;
	int cnt;
	GL_ERROR_MSG("MBAScene::drawGL -- enter");
	for (i = 0; i < (int)lightGroupList->count(); i++)
		lightGroupList->at(i)->illuminateGL(frame);
	GL_ERROR_MSG("MBAScene::drawGL -- illuminate");
	
	for (i = 0; i < (int)lightGroupList->count(); i++)
		lightGroupList->at(i)->drawGL(frame);
	GL_ERROR_MSG("MBAScene::drawGL -- draw lights");
	
	for (i = 0; i < (cnt = (int)particleList->count()); i++)
	{
		particleList->at(i)->drawGL(frame);
	}
	GL_ERROR_MSG("MBAScene::drawGL -- particles");
}

//--------------------------------------------------------------------
bool	MBAScene::GotoFrame(int newFrame)
{
	int		i;
	int		fdiff;
	char	labelText[128];
	bool	completed = true;
	
	if (newFrame < g_frame)
	{
		reset();
		fdiff = newFrame;
	}
	else if (newFrame > g_frame)
		fdiff = newFrame - g_frame;
	else
		fdiff = 0;

	QProgressDialog progress( "Calculating frame...", "Cancel", fdiff, 0, "frameProgress", true );
	progress.setIcon(QPixmap((const char**)flowIcon_xpm));
	progress.setCaption("Calculating Frame");
	progress.setMinimumDuration(2);
	for (i = 0; i < fdiff; i++)
	{
		progress.setProgress( i );
		sprintf(labelText, "Calculating frame...%d", g_frame);
		progress.setLabelText( labelText );
		if ( progress.wasCancelled() )
		{
			completed = false;
			break;
		}
		execute(g_frame);	
		g_frame++;
	}
    progress.setProgress( fdiff );
	return completed;	
}

//--------------------------------------------------------------------
void	MBAScene::AddLightGroup(LightGroup* lg, int index)
{
	if (lg)
	{
		if (index >= 0)
		{
			if (index <= (int)lightGroupList->count())
				lightGroupList->insert((uint)index, lg);
			else
			{
				warning("MBAScene::AddLightGroup(%p, %d) - index out of range. Appending.", lg, index);
				lightGroupList->append(lg);
			}
		}
		else
			lightGroupList->append(lg);
	}
	if(lightGroupList->count() == 1)
	{
		activeLightGroup = 0;
	}
}

//--------------------------------------------------------------------
void	MBAScene::RemoveLightGroup(int index)
{
	if (index >= 0 && index < (int)lightGroupList->count())
	{
		lightGroupList->remove(index);
	}
	if(activeLightGroup == index)
	{
		if(lightGroupList->count() >= 1)
			activeLightGroup = 0;
		else
			activeLightGroup = -1;
	}
}

//--------------------------------------------------------------------
void	MBAScene::AddParticleSystem(ParticleControl* pc, int index)
{
	if (pc)
	{
		if (index >= 0)
		{
			if (index <= (int)particleList->count())	
				particleList->insert((uint)index, pc);
			else
			{
				warning("MBAScene::AddLightGroup(%p, %d) - index out of range. Appending.", pc, index);
				particleList->append(pc);
			}
		}
		else
			particleList->append(pc);
	}
}
	
//--------------------------------------------------------------------
void	MBAScene::RemoveParticleSystem(int index)
{
	if (index >= 0 && index < (int)particleList->count())
	{
		particleList->remove(index);
	}
}

//--------------------------------------------------------------------
LightGroup* MBAScene::LightGroupAt(int index)
{
	if (index < (int)lightGroupList->count())
		return lightGroupList->at(index);
	else
	{
		warning("MBAScene ERROR: request for lightGroup at %d - count is %d.", index, (int)lightGroupList->count() );
		return 0;
	}
}

//--------------------------------------------------------------------
CameraView* MBAScene::CameraAt(int index)
{
	if (index < (int)cameraList->count())
		return cameraList->at(index);
	else
	{
		warning("MBAScene ERROR: request for camera at %d - count is %d.", index, (int)cameraList->count() );
		return 0;
	}
}

//--------------------------------------------------------------------
ParticleControl* MBAScene::ParticleSystemAt(int index)
{
	if (index < (int)particleList->count())
		return particleList->at(index);
	else
	{
		warning("MBAScene ERROR: request for particle system at %d - count is %d.", index, (int)particleList->count() );
		return 0;
	}
}

//--------------------------------------------------------------------
LightGroup* MBAScene::ActiveLightGroup()
{
	if(activeLightGroup >= 0)
		if(activeLightGroup < (int)lightGroupList->count())
		{
			return lightGroupList->at(activeLightGroup);
		}
		else
			warning("ERROR:MBAScene::ActiveLightGroup() activeLightGroup > lightGroupList.count()");
	return 0;
}

//--------------------------------------------------------------------
CameraView* MBAScene::ActiveCamera()
{
	if(activeCamera >= 0)
		if(activeCamera < (int)cameraList->count())
		{
			return cameraList->at(activeCamera);
		}
		else
			warning("ERROR: MBAScene::ActiveCamera() activeCamera > (int)cameraList.count()");
	return 0;
}

//--------------------------------------------------------------------
int MBAScene::setActiveParticleSystem(int index)
{
	if(index >= 0 && index < (int)particleList->count())
		activeParticleSystem = index;
	return activeParticleSystem;
}

//--------------------------------------------------------------------
ParticleControl* MBAScene::ActiveParticleSystem()
{
	if(activeParticleSystem >= 0)
		if(activeParticleSystem < (int)particleList->count())
			return particleList->at(activeParticleSystem);
	return 0;
}

//--------------------------------------------------------------------
void	MBAScene::ResetParticleSystems()
{
	for(int i = 0; i < numParticleSystems(); i++)
		ParticleSystemAt(i)->ResetSystem();
}

//--------------------------------------------------------------------
void	MBAScene::saveToFile(FILE* outfile)
{
	int i;
	
	fprintf(outfile, "RandomSeed      %d\n", g_randSeed );
	fprintf(outfile, "RenderName      %s\n", g_rendName );
	fprintf(outfile, "RenderExecOpts  %s\n", g_rendermanOpts );
	fprintf(outfile, "CodeConfig      %s\n", (const char*)codeConfigFile);
	fprintf(outfile, "BackgroundColor %f %f %f\n", backgroundColor->R(), backgroundColor->G(), backgroundColor->B() );
	
	fprintf(outfile, "CameraPos       %f %f %f\n", g_globalCam->GetVals().position.X(), 
											g_globalCam->GetVals().position.Y(),
											g_globalCam->GetVals().position.Z() );
	fprintf(outfile, "CameraFOV       %f\n", g_globalCam->GetVals().fov );
	fprintf(outfile, "TargetPos       %f %f %f\n", g_globalCam->GetVals().target.X(), 
											g_globalCam->GetVals().target.Y(),
											g_globalCam->GetVals().target.Z() );
	
	fprintf(outfile, "glCullFace      %d\n", g_glCullFace);
	fprintf(outfile, "glBlend         %d\n", g_glBlend);
	fprintf(outfile, "glTwoSideLight  %d\n", g_glTwoSideLight);
	

	for (i = 0; i < (int)lightGroupList->count(); i++)
		lightGroupList->at(i)->SaveLightGroup(outfile, 0);
	
	for (i = 0; i < (int)particleList->count(); i++)
		particleList->at(i)->SaveParticleSystem(outfile, 0);
}

//--------------------------------------------------------------------
void	MBAScene::loadFromFile(FILE* infile)
{
	if(!infile)
		return;

	delete [] g_rendName;
	g_rendName = new char[5];
	strcpy(g_rendName, "flow");
	
	GLState newGLState = g_camGLPort->getGLState();
	
	int		i = 0;
	float	r,g,b;
	float	x,y,z;
	char	buffer[512];
	int		trunc;
	QCString	keyString;
	QCString	valString;
	while (fgets(buffer, 512, infile) > 0)
	{
		keyString = buffer;
		keyString = keyString.simplifyWhiteSpace();

		valString = (const char*)keyString;
		trunc = valString.find(' ');
		if(trunc < 0)
			valString = "";
		else
			valString.remove(0, trunc+1);
	
		if ( keyString.find('#', 0) == 0)
			continue; // ignore comment
		else if ( keyString.find("RenderName", 0, true) == 0 )
		{
			delete [] g_rendName;
			g_rendName = new char[strlen((const char*)valString)+1];
			strcpy(g_rendName, (const char*)valString);
		}
		else if ( keyString.find("RandomSeed", 0, true) == 0 )
		{
			g_randSeed = valString.toInt();
		}
		else if ( keyString.find("BackgroundColor", 0, true) == 0 )
		{
			sscanf(valString, "%f %f %f", &r, &g, &b);
			this->setBackgroundColor(r, g, b);
		}
		else if ( keyString.find("RenderExecOpts", 0, true) == 0 )
		{
			if(g_rendermanOpts)
				delete [] g_rendermanOpts;
			g_rendermanOpts = new char[strlen((const char*)valString)+1];
			strcpy(g_rendermanOpts, (const char*)valString);
		}
		else if ( keyString.find("CameraPos", 0, true) == 0 )
		{
			sscanf(valString, "%f %f %f", &x, &y, &z);
			g_globalCam->SetTrans(x, y, z);
		}
		else if ( keyString.find("CameraFOV", 0, true) == 0 )
		{
			sscanf(valString, "%f", &x);
			g_globalCam->SetFov(x);
		}
		else if ( keyString.find("TargetPos", 0, true) == 0 )
		{
			sscanf(valString, "%f %f %f", &x, &y, &z);
			g_globalCam->SetTarget(x, y, z);
		}
		else if ( keyString.find("ParticleSystem", 0, true) == 0 )
		{
			warning("---- Loading %-20s : %s", "Particle System", (const char*)valString);
			//g_particleControlWidget->SetParticleControl((ParticleControl*)0);
			ParticleControl	*tmpControl;
			tmpControl = new ParticleControl();
			tmpControl->LoadParticleSystem(infile, (const char*)valString);
			this->AddParticleSystem(tmpControl);
		}
		else if (keyString.find("LightGroup", 0, true) == 0 )
		{
			warning("---- Loading %-20s : %s", "Light Group", (const char*)valString);
			LightGroup	*tmpLightGroup = new LightGroup();
			tmpLightGroup->LoadLightGroup(infile, (const char*)valString);				
			this->AddLightGroup(tmpLightGroup);
		}
		else if (keyString.find("glCullFace", 0, true) == 0 )
			g_glCullFace = valString.toInt();	
		else if (keyString.find("glBlend", 0, true) == 0 )
			g_glBlend = valString.toInt();	
		else if (keyString.find("glTwoSideLight", 0, true) == 0 )
			g_glTwoSideLight = valString.toInt();	
		else if (keyString.find("CodeConfig", 0, true) == 0 )
		{
			codeConfigFile = valString;
			loadCodeConfig(valString); 
		}
		else
		{
			warning ("MBAScene::loadFromFile -- %04d : \'%s\'", i++, (const char*)keyString);
		}
	}
	newGLState.setCullFace(g_glCullFace);
	newGLState.setBlend(g_glBlend);
	newGLState.setTwoSideLight(g_glTwoSideLight);
	g_camGLPort->setGLState(newGLState);

}

//--------------------------------------------------------------------
bool MBAScene::loadCodeConfig(const QString &filename)
{
	bool		retVal = false;
	QFile		infile;
	QString		lineString;
	int			tmp;
	
	if(!filename.isEmpty())
	{
		infile.setName(filename);
		if(infile.open(IO_ReadOnly))
		{
			codeConfigFile = filename;
			codeFiles.clear();
			while( infile.readLine(lineString, 512) >= 0 )
			{
				if( (tmp=lineString.find("SCAN")) >= 0)
				{
					lineString.remove(tmp, 4);
					lineString = lineString.simplifyWhiteSpace();
					codeFiles.append(lineString);
				}
				if( (tmp=lineString.find("OUTPUT")) >= 0)
				{
					lineString.remove(tmp, 6);
					lineString = lineString.simplifyWhiteSpace();
					codeOutDir = lineString;
				}
			}
			codeFiles.sort();
			infile.close();
			retVal = true;
		}
	}
	return retVal;
}

//--------------------------------------------------------------------
bool MBAScene::saveCodeConfig(const QString &filename)
{
	bool	retVal = false;
	QFile	outFile(filename);
	
	if(outFile.open(IO_WriteOnly))
	{
		codeConfigFile = filename;
		QTextStream	textStream(&outFile);
		
		for(QStringList::Iterator it = codeFiles.begin(); it != codeFiles.end(); ++it )
		{
			textStream << "SCAN      " << (*it).latin1() << '\n';
		}
		textStream << "OUTPUT    " << codeOutDir << '\n';
		outFile.close();
		retVal = true;
	}
	return retVal;
}


