/**
 * mbaScene.cpp
 * Copyright 1999 by Mark B. Allan
 * 
 * contains and controls the scene
 */
#include "mbaScene.h"

#include <qprogressdialog.h>

#include "extern.h"
#include "define.h"
#include "lightGroup.h"
#include "particleControl.h"
#include "camera.h"

enum {	FLOW_VARIABLE_DECLARE = 0, 
		FLOW_FUNCTION_DECLARE,
		FLOW_VARIABLE_INIT, 
		FLOW_EXEC,
		FLOW_FUNCTIONS
};

//--------------------------------------------------------------------
bool	MBAScene::writeCode()
{
	char	buffer[1024];
	char	*walker  = NULL;
	FILE	*outfile = NULL;
	FILE	*templateFile = NULL;
	QString	templateFileName;
	QString	outfileName;
	
	const char	*flowTag[5] = {	"@FLOW_VARIABLE_DECLARE",
								"@FLOW_FUNCTION_DECLARE",
								"@FLOW_VARIABLE_INIT",
								"@FLOW_EXEC",
								"@FLOW_FUNCTIONS" };
		
	int	count = (int)codeFiles.count();
	QProgressDialog	progress("writing files...", "cancel", count, 0, 0, true);
	
	int i = 0;
	QStringList::Iterator it; 
	for(it = codeFiles.begin(); it != codeFiles.end(); it++)
	{
		progress.setProgress(i++);
		if(progress.wasCancelled())
			return false;
					
		templateFileName = (const char*)(*it);
		outfileName = (const char*)(*it);
		int tmp = outfileName.findRev('/');
		if(tmp >= 0)
			outfileName.remove(0, tmp);
		outfileName.insert(0, codeOutDir);
		
		if( templateFileName.compare(outfileName) == 0)
		{
			warning("ERROR: Template file and output file are the same! (%s)", (const char*)templateFileName);
			return false;
		}
		
		templateFile = fopen(templateFileName, "r");
		if(!templateFile)
		{
			warning("ERROR: MBAScene::writeCode() -- could not open %s for read\n", (const char*)templateFileName);
			return false;
		}
		outfile = fopen(outfileName, "w");
		if(!outfile)
		{
			warning("ERROR: MBAScene::writeCode() -- could not open %s for write\n", (const char*)outfileName);
			return false;
		}
		
		//-- Copy files to new location, scanning for flow tags...
		while(fgets(buffer, 1024, templateFile))
		{
			if( (walker = strchr(buffer, '/')) && *(walker+1) == '/' )
			{
				fputs(buffer, outfile);
			}
			else if( (walker = strchr(buffer, '@')) )	
			{
				if( strncmp(walker, flowTag[FLOW_VARIABLE_DECLARE], strlen(flowTag[FLOW_VARIABLE_DECLARE])) == 0 )
				{
					writeCode_VARIABLE_DECLARE(outfile);
				}
				else if( strncmp(walker, flowTag[FLOW_FUNCTION_DECLARE], strlen(flowTag[FLOW_FUNCTION_DECLARE])) == 0 )
				{
					writeCode_FUNCTION_DECLARE(outfile);
				}
				else if(strncmp(walker, flowTag[FLOW_VARIABLE_INIT], strlen(flowTag[FLOW_VARIABLE_INIT])) == 0 )
				{
					writeCode_VARIABLE_INIT(outfile);
				}
				else if( strncmp(walker, flowTag[FLOW_EXEC], strlen(flowTag[FLOW_EXEC])) == 0 )	
				{
					writeCode_EXEC(outfile);
				}
				else if( strncmp(walker, flowTag[FLOW_FUNCTIONS], strlen(flowTag[FLOW_FUNCTIONS])) == 0 )	
				{
					writeCode_FUNCTIONS(outfile);
				}
				else
					fputs(buffer, outfile); // didn't find anything, so copy to file
			}
			else
			{
				fputs(buffer, outfile);	// default to copying line-by-line
			}
		}		
		fclose(templateFile);
		fclose(outfile);
	}
	return true;
}


//--------------------------------------------------------------------
void	MBAScene::writeCode_VARIABLE_DECLARE (FILE *outfile)
{
	CameraVals	curVals (g_globalCam->GetVals());
	int 		numGroups = (int)particleList->count();
	if(outfile)
	{
		fprintf(outfile, "/*-- FLOW_VARIABLE_DECLARE --------*/\n");
		fprintf(outfile, "	#define FLOW_NUM_GROUPS %d \n", numGroups);
		fprintf(outfile, "	int    particle_handle[FLOW_NUM_GROUPS];\n");
		fprintf(outfile, "	int    maxParticles[FLOW_NUM_GROUPS];   \n");
		fprintf(outfile, "	float  fov;                \n");
		fprintf(outfile, "	float  nearPlane;          \n");
		fprintf(outfile, "	float  farPlane;           \n");
		fprintf(outfile, "	MMatrix camMat;            \n");
		fprintf(outfile, "	MVertex camPos;            \n");
		fprintf(outfile, "	double  rateAdj;           \n");
		fprintf(outfile, "	double  timeAdj;           \n");
		fprintf(outfile, "	int  	frame;             \n");
	}
}

//--------------------------------------------------------------------
void	MBAScene::writeCode_FUNCTION_DECLARE (FILE *outfile)
{
	if(outfile)
	{
		fprintf(outfile, "/*-- FLOW_FUNCTION_DECLARE -------*/\n");
		for (int i = 0; i < (int)particleList->count(); i++)
			fprintf(outfile, "	void	execParticleGroup%02d();\n", i);

	}
}

//--------------------------------------------------------------------
void	MBAScene::writeCode_VARIABLE_INIT (FILE *outfile)
{
	CameraVals	curVals (g_globalCam->GetVals());
	if(outfile)
	{
		fprintf(outfile, "/*-- FLOW_VARIABLE_INIT --------*/\n");
		for (int i = 0; i < (int)particleList->count(); i++)
		{
			fprintf(outfile, "	maxParticles[%d]    = %d;\n", i, particleList->at(i)->getMaxParticles() );
			fprintf(outfile, "	particle_handle[%d] = pGenParticleGroups(1, maxParticles[%d]);\n", i, i);
		}
		fprintf(outfile, "	fov 	  = %f;\n", curVals.fov);
		fprintf(outfile, "	nearPlane = %f;\n", g_globalCam->getNear());
		fprintf(outfile, "	farPlane  = %f;\n", g_globalCam->getFar());
		fprintf(outfile, "	camMat.setTranslate( %f, %f, %f );\n", -curVals.position.X(), -curVals.position.Y(), -curVals.position.Z() );
		fprintf(outfile, "	camMat.rotateY ( %f );\n", curVals.rotate.Y() );
		fprintf(outfile, "	camMat.rotateX ( %f );\n", curVals.rotate.X() );
		fprintf(outfile, "	camMat.rotateZ ( %f );\n", curVals.rotate.Z() );
		fprintf(outfile, "	camPos.setX( %f );\n", curVals.position.X() );
		fprintf(outfile, "	camPos.setY( %f );\n", curVals.position.Y() );
		fprintf(outfile, "	camPos.setZ( %f );\n", curVals.position.Z() );
		fprintf(outfile, "	rateAdj = 1.0;\n");
		fprintf(outfile, "	timeAdj = 1.0;\n");
		fprintf(outfile, "	frame   = 0;\n");
		fprintf(outfile, "\n");
	}
}

//--------------------------------------------------------------------
void	MBAScene::writeCode_EXEC (FILE *outfile)
{
	if(outfile)
	{
		fprintf(outfile, "/*-- FLOW_EXEC --------*/\n");
		for (int i = 0; i < (int)particleList->count(); i++)
		{
			if(particleList->at(i)->isEnabled())
				fprintf(outfile, "	flow->execParticleGroup%02d();\n", i);
			else
				fprintf(outfile, "	/*flow->execParticleGroup%02d();*/\n", i);
				
			switch (particleList->at(i)->PrimitiveType())
			{
				case ParticleControl::Point:
					//fprintf(outfile, "	glPointSize(%f);\n", particleList->at(i)->PointSize() );
					fputs ("	glDraw.pDrawGroupp(Point);\n", outfile);
					break;
				case ParticleControl::Line:
					//fprintf(outfile, "	glLineWidth(%f);\n", particleList->at(i)->LineWidth() );
					fprintf(outfile, "	glDraw.pDrawGroupp(Line);\n");
					break;
				case ParticleControl::Polygon:
					break;
				case ParticleControl::Disk:
					fprintf(outfile, "	glDraw.pDrawGroupp(Disk);\n");
					break;
				case ParticleControl::Connect:
					break;
			}
		}
	
	}
}

//--------------------------------------------------------------------
void	MBAScene::writeCode_FUNCTIONS (FILE *outfile)
{
	if(outfile)
	{
		fprintf(outfile, "/*-- FLOW_FUNCTIONS --------*/\n");
		for (int i = 0; i < (int)particleList->count(); i++)
		{
			fprintf(outfile, "/*------------------------------------------------------------------*/\n");
			fprintf(outfile, "void Flow::execParticleGroup%02d()\n", i);
			fprintf(outfile, "{\n");
			fprintf(outfile, "	pCurrentGroup(particle_handle[%d]);\n", i);
			particleList->at(i)->WriteCode(outfile);
			fprintf(outfile, "	pMove();\n");
			fprintf(outfile, "}\n\n");
		}
	}
}

