/**
 * numInput.cpp
 * Copyright 1999 by Mark B. Allan
 * 
 * numerical input w/ 'spinbox' label
 */
#include "numInput.h"
#include "numInputValid.h"
#include <limits.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif

#include <qevent.h>
#include <qapp.h>
#include <qclipboard.h>
#include <qtimer.h>
#include <qcursor.h>
 
#include "extern.h" 

//=======================================================================
//											NumInput
//=======================================================================
NumInput::NumInput ( int t, QWidget* parent, const char* name )
	: QWidget (parent, name)
{
	labelWidth		= 20;
	height			= 20;
	editMinWidth	= 20;
	editWidth		= 50;
	buttonMode		= false;
	
	currentValue	= 0.0000001;
	currentIntValue	= 0;
	
	inputLabel	= new NumInputLabel(this);
	inputLabel->setCursor (sizeHorCursor);
	inputLabel->setText("-");
	height = inputLabel->sizeHint().height();
	inputLabel->setFixedHeight(height);
	inputLabel->setFixedWidth(labelWidth);
	inputEdit	= new NumInputEdit(this);
	inputEdit->setFixedHeight(height);
	inputEdit->setMinimumWidth(editMinWidth);
	inputEdit->setMaxLength(8);
	
	inputButton = new MToolButton(this);
	inputButton->setFixedHeight(height);
	inputButton->setMinimumWidth(editMinWidth);
	inputButton->hide();
	
	// I think it was this that was causing the ASSERT:XGetSelectionOwner error:
	//setFocusProxy(inputEdit);
	//setFocusPolicy(QWidget::StrongFocus);
		
	inputLabel->move(0, 0);
	inputEdit->move(labelWidth, 0);
		
	setFixedHeight(height);
	SetType (t);
	setEnabled(true);

}

NumInput::~NumInput()
{
}

//-----------------------------------------------------------------------
void	NumInput::resizeEvent(QResizeEvent *rEv)	
{
	QSize newSize = rEv->size();
	int tmpW = newSize.width()-labelWidth;
	if(tmpW > editMinWidth)
	{
		editWidth = tmpW;
		inputEdit->resize(editWidth, height);
		inputButton->resize(editWidth, height);
	}
}

//-----------------------------------------------------------------------
void	NumInput::SetFocus ()	
{ 
	if (hasFocus())
	{
		inputEdit->selectAll();
	}
	else
	{
		inputEdit->selectAll();
		inputEdit->setFocus();
	}
		
}

//-----------------------------------------------------------------------
void	NumInput::SetHeight (int h)	
{ 
	setFixedHeight(h);
	inputLabel->setFixedHeight(h);
	inputEdit->setFixedHeight(h);
	inputButton->setFixedHeight(h);
}

//-----------------------------------------------------------------------
void	NumInput::SetLabelWidth (int w)	
{ 
	labelWidth = w; 
	inputLabel->setFixedWidth(labelWidth); 
	inputEdit->move(labelWidth, 0);
	inputButton->move(labelWidth, 0);
}

///-----------------------------------------------------------------------
void	NumInput::SetTotalWidth (int w)	
{ 
	editMinWidth = editWidth = w-2-labelWidth; 
	inputEdit->setMinimumSize(editMinWidth, height);
	inputEdit->resize(editWidth, height);
	inputButton->setMinimumSize(editMinWidth, height);
	inputButton->resize(editWidth, height);
	setMinimumWidth(w); 
}

//-----------------------------------------------------------------------
void	NumInput::setEnabled (bool status)
{
	active = status;
	inputLabel->SetActive(status);
	inputEdit->SetActive(status);
	inputButton->setEnabled(status);
	if (status)
	{
		switch(type)
		{
			case NumInput::Double:
				inputEdit->SetValue(currentValue);
				break;
			case NumInput::Int:
				inputEdit->SetValue(currentIntValue);
				break;
		}
	}
}

//-----------------------------------------------------------------------
void	NumInput::toggleButtonMode()
{
	buttonMode = !buttonMode;
	setButtonMode(buttonMode);
}

//-----------------------------------------------------------------------
void	NumInput::setButtonMode (bool status)
{
	buttonMode = status;
	if(buttonMode)
	{
		inputEdit->hide();
		inputButton->show();
	}
	else
	{
		inputEdit->show();
		inputButton->hide();
	}
}
	
//-----------------------------------------------------------------------
void	NumInput::SetLabel (const char *text)
{
	inputLabel->setText(text);
	inputLabel->lastText = text;
}

//-----------------------------------------------------------------------
void	NumInput::SetType (int t)
{
	type = t;
	inputEdit->SetType(t);
	inputLabel->SetType(t);
}

//-----------------------------------------------------------------------
float	NumInput::SetValue (float in)
{
	if ( (type == NumInput::Double) && (in != (float)currentValue) )
	{
		currentValue	= (double)in;
		currentIntValue	= (int) in;
		if (active)
			inputEdit->SetValue (currentValue);
		emit ValueChanged (currentValue);
		emit ValueChanged ((float)currentValue);
	}
	else if ( (type == NumInput::Int) && ((int)in != currentIntValue) )
	{
		currentValue	= (double)in;
		currentIntValue	= (int) in;
		if (active)
			inputEdit->SetValue (currentIntValue);
		emit ValueChanged(currentIntValue);
	}
	return (float)currentValue;
}

//-----------------------------------------------------------------------
float	NumInput::SetValueSolid (float in)
{
	if ( (type == NumInput::Double) )
	{
		currentValue	= (double)in;
		currentIntValue	= (int) in;
		if (active)
			inputEdit->SetValue (currentValue);
	}
	else if ( (type == NumInput::Int) )
	{
		currentValue	= (double)in;
		currentIntValue	= (int) in;
		if (active)
			inputEdit->SetValue (currentIntValue);
	}
	return (float)currentValue;
}

//-----------------------------------------------------------------------
double	NumInput::SetValue (double in)
{
	if ( (type == NumInput::Double) && (in != currentValue) )
	{
		currentValue	= in;
		currentIntValue	= (int) in;
		inputEdit->SetValue (currentValue);
		emit ValueChanged (currentValue);
		emit ValueChanged ((float)currentValue);
	}
	else if ( (type == NumInput::Int) && ((int)in != currentIntValue) )
	{
		currentValue	= in;
		currentIntValue	= (int) in;
		inputEdit->SetValue (currentIntValue);
		emit ValueChanged(currentIntValue);
	}
	return currentValue;
}

double	NumInput::SetValueSolid (double in)
{
	if ( type == NumInput::Double )
	{
		currentValue	= in;
		currentIntValue	= (int) in;
		inputEdit->SetValue (currentValue);
	}
	else if ( type == NumInput::Int )
	{
		currentValue	= in;
		currentIntValue	= (int) in;
		inputEdit->SetValue (currentIntValue);
	}
	return currentValue;
}

//-----------------------------------------------------------------------
int		NumInput::SetValue (int in)
{
	if ( (type == NumInput::Double) && (in != currentValue) )
	{
		currentValue	= (double) in;
		currentIntValue	= (int) in;
		inputEdit->SetValue (currentValue);
		emit ValueChanged (currentValue);
		emit ValueChanged ((float)currentValue);
	}
	else
	if ( (type == NumInput::Int) && ((int)in != currentIntValue) )
	{
		currentValue	= (double)in;
		currentIntValue	= (int) in;
		inputEdit->SetValue (currentIntValue);
		emit ValueChanged(currentIntValue);
	}
	return currentIntValue;
}

int		NumInput::SetValueSolid (int in)
{
	if ( type == NumInput::Double )
	{
		currentValue	= (double) in;
		currentIntValue	= (int) in;
		inputEdit->SetValue (currentValue);
	}
	else
	if ( type == NumInput::Int )
	{
		currentValue	= (double)in;
		currentIntValue	= (int) in;
		inputEdit->SetValue (currentIntValue);
	}
	return currentIntValue;
}

//-----------------------------------------------------------------------
void	NumInput::add (double in)
{
	currentValue	+=	in;
	
	if ( (type == NumInput::Int) && (currentIntValue != (int)currentValue) )
	{
		currentIntValue = (int)currentValue;
		emit ValueChanged (currentIntValue);
		inputEdit->SetValue(currentIntValue);	
	}
	else if ( type == NumInput::Double)
	{
		emit ValueChanged ((float)currentValue);
		inputEdit->SetValue(currentValue);	
	}
}
		
//-----------------------------------------------------------------------
void	NumInput::InputComplete ()
{
	emit Updated();
}

//=======================================================================
//											NumInput Label
//=======================================================================
NumInputLabel::NumInputLabel (NumInput* p, const char* name)
	: QLabel ( p, name )
{
	numInput 	= p;
	moved		= 0;
	triggered	= false;
	
	ctrlKey		= 0;
	altKey		= 0;
	shftKey		= 0;
	
	setFont( *g_font_SB );
	setAlignment ( AlignCenter );
	setFrameStyle ( Panel | Sunken );
	setFocusPolicy (QWidget::TabFocus);
	
	setBackgroundMode(PaletteButton);
//	setFocusPolicy (QWidget::NoFocus);
}

NumInputLabel::~NumInputLabel ()
{
}
	
//-----------------------------------------------------------------------
//FIXME -  THIS IS A HORRIBLE, UGLY HACK TO GET AROUND THE X ERRORS I WAS
//FIXME -  GETTING FROM selectAll() in NumInputEdit::focusInEvent(). THIS
//FIXME -  PREVENTS Shift-Tab FROM MOVING FOCUS BACKWARDS. BUT I LIKE AUTOSELECT
//FIXME -  FROM TAB - MAKES NUMBER INPUT MUCH FASTER
void	NumInputLabel::focusInEvent ( QFocusEvent * )
{
	numInput->SetFocus();
}
//void	NumInputLabel::focusInEvent ( QFocusEvent * )
//{
//}

//-----------------------------------------------------------------------
void	NumInputLabel::SetActive ( bool status )
{
	active = status;
	
	setEnabled (active);
	if (active)
	{
		setText(lastText);
	}
	else
	{
		if(strlen(text()) > 0)
			lastText = text();
		setText("");
	}
}
	
//-----------------------------------------------------------------------
void	NumInputLabel::mousePressEvent  (QMouseEvent *msEv)
{
//	int	but;
//	but = msEv->button();
	lastXpos = msEv->x();
	triggered = false;
	moved = 0;
	
	if(msEv->button() == RightButton)
		numInput->toggleButtonMode();
		
//	xOrig =  msEv->x();
//	mousePos = QCursor::pos();
}
		
//--------------------------------------------------------------------
void	NumInputLabel::mouseReleaseEvent  (QMouseEvent * msEv)
{
	if (type == NumInput::Int)
		( (NumInput*)parent() )->IntifyCurrentValue();
	if (!moved && msEv->button() == LeftButton)
	{
//		numInput->setFocus();
		numInput->SetFocus();
	}
}	

//--------------------------------------------------------------------
void	NumInputLabel::mouseMoveEvent  (QMouseEvent *msEv)
{
//	static	int	count = 0;
	double	inc = 0;
	
	if (!triggered)	// mechanism to allow some slop with the mouseclick
	{
		if ( abs(msEv->x()-lastXpos) > 6 )
		{
			triggered = true;
			lastXpos = msEv->x();
		}
		else
			return;
	}
		
	moved = msEv->x() - lastXpos;
	//warning ("moved = %d, active %d, type = %d", moved, active, type);

	if (active)
	{
		switch (type)
		{
//			//-----------------
//			case NumInput::Int:
//				switch (msEv->state())
//				{
//					case (RightButton):
//						break;
//				}
//				break;
			//-----------------
			case NumInput::Double:
				switch (msEv->state())
				{
					case (NoButton):
						break;
					case (LeftButton):
						inc = (float)moved * 0.100f;
						break;
					case (LeftButton | AltButton):
						inc = (float)moved * 0.010f;
						break;
					case (LeftButton | AltButton | ShiftButton):
						inc = (float)moved * 0.001f;
						break;
					case (LeftButton | ControlButton):
						inc = (float)moved * 1.00f;
						break;
					case (LeftButton | ControlButton | ShiftButton ):
						inc = (float)moved * 10.0f;
						break;
					case (LeftButton | ControlButton | ShiftButton | AltButton ):
						inc = (float)moved * 100.0f;
						break;
					case (LeftButton | ShiftButton):
						inc = (float)moved * 0.100f;
						break;
					default:
						break;
				}
				break;
			//-----------------
			default:
				warning ("\aNumInputLabel::mouseMoveEvent() - invalid type");
				break;
		}
		( (NumInput*)parent() )->add (inc);
	}
	
	lastXpos = msEv->x();
	( (NumInput*)parent() )->InputComplete();
//	if( (count++%5) == 0)
//	{
//		QCursor::setPos(mousePos);
//		lastXpos = xOrig;
//	}
}
	
//=======================================================================
//											NumInput Line Edit
//=======================================================================
NumInputEdit::NumInputEdit ( QWidget* parent, const char* name )
	: QLineEdit ( parent, name )
{
	lastGoodValue = 0.0;

	setEchoMode( QLineEdit::Normal );
	setFrame( true );
	
	setFont( *g_font_M );
//	setPalette(*g_litePalette);
	
	type = NumInput::Double;
	currentValidator = new NumInputDblValidator (this);
//	currentValidator = new QDoubleValidator(-200000.0, 200000, 10, this);
	setValidator ( currentValidator );	
	setText( "0" );

	QObject::connect (this, SIGNAL(returnPressed()), SLOT(InputComplete()) );
}

NumInputEdit::~NumInputEdit()
{
}

//-----------------------------------------------------------------------
void	NumInputEdit::focusInEvent ( QFocusEvent *fE )
{
	QLineEdit::focusInEvent(fE);
//	qApp->syncX();
//	qApp->flushX();
//	QTimer::singleShot(75, this, SLOT(selectAll()) );
//	selectAll();
//	copy();
}

//-----------------------------------------------------------------------
void	NumInputEdit::focusOutEvent ( QFocusEvent *fE )
{
//	deselect();
	InputComplete();
	QLineEdit::focusOutEvent(fE);
}

//-----------------------------------------------------------------------
void	NumInputEdit::SetActive (bool status)
{
	QCString tempString;
	
	active = status;
	if (status)
	{
//		setEnabled (TRUE);
		if (type == NumInput::Double)
			tempString.setNum(lastGoodValue);
		if (type == NumInput::Int)
			tempString.setNum((int)lastGoodValue);
		setText(tempString);
	}
	else
	{
		setText ("-");
	}	
	setEnabled (status);
}

//-----------------------------------------------------------------------
void	NumInputEdit::InputComplete ()
{
	QCString	tempString1;
	QCString	tempString2;
	double	tmpDbl;
		
	tempString1 = text();
	if (type == NumInput::Double)
	{
		tmpDbl = tempString1.toDouble();
		tempString2.setNum(tmpDbl);
		lastGoodValue = tmpDbl;
		( (NumInput*)parent() )->SetValue(lastGoodValue);
		( (NumInput*)parent() )->InputComplete();
	}
	else if (type == NumInput::Int)
	{
		tmpDbl = tempString1.toDouble();
		tempString2.setNum((int)tmpDbl);
		lastGoodValue = (double)(int)tmpDbl;
		( (NumInput*)parent() )->SetValue(lastGoodValue);
		( (NumInput*)parent() )->InputComplete();
	}
	else 
	{
		warning ("\aNumInput::InputComplete() - bad type");
	}
	setText (tempString2);
	clearFocus();
	
//	warning ("NumInputEdit::InputComplete - lastGoodValue = %f", lastGoodValue);
}
	
double	NumInputEdit::GetLastGoodValue ()
{
	return lastGoodValue;
}

//-----------------------------------------------------------------------
void	NumInputEdit::SetType (int t)
{
	if (t == type)
		return;
		
	type = t;
	if (type == NumInput::Double)
	{
		setValidator (0);
		delete currentValidator;
		//currentValidator = new QDoubleValidator(-(INT_MAX-1), (INT_MAX-1), 6, this);
		currentValidator = new NumInputDblValidator (this);
		setValidator ( currentValidator );
	}
	else if (type == NumInput::Int)
	{
		setValidator (0);
		delete currentValidator;
		currentValidator = new NumInputIntValidator (this);
		//currentValidator = new QIntValidator(-(INT_MAX-1), (INT_MAX-1), this);
		setValidator ( currentValidator );
	}
	else
	{
		warning ("\aNumInputEdit::SetType() - Invalid type");
	}
}	

//-----------------------------------------------------------------------
void	NumInputEdit::SetValue (int in)
{
	QCString	tmp;
	tmp.setNum(in);
	setText (tmp);
}
void	NumInputEdit::SetValue (double in)
{
	QCString	tmp;
	tmp.setNum(in);
	setText (tmp);
}
void	NumInputEdit::SetValue (float in)
{
	QCString	tmp;
	tmp.setNum(in);
	setText (tmp);
}

