#include "shaderPreviewList.h"
#include "Pixmaps/noPreview.xpm"

#include <stdlib.h>
#include <stdio.h>

#include <qpainter.h>
#include <qkeycode.h>
#include <qprinter.h>
#include <qpen.h> 

#include <qpixmapcache.h> 

extern int g_previewTableCellSize;

//==============================================================================
ShaderPreview::ShaderPreview(int numCols, QWidget *parent, const char *name )
    : QTableView(parent,name)
{
	curRow	= curCol	= 0;						// set currently selected cell
	lastRow	= lastCol	= 0;						// set currently selected cell
	enabled = true;
	
	setFocusPolicy( StrongFocus );				// we accept keyboard focus
	maxCols = numCols;
	setNumCols( maxCols );						// set number of col's in table
	setNumRows( 0 );							// set number of rows in table
	setCellWidth ( g_previewTableCellSize+6 );	// set width of cell in pixels
	setCellHeight( g_previewTableCellSize+6 );	// set height of cell in pixels
	setTableFlags(	Tbl_autoVScrollBar |		// auto scrollbar
					Tbl_clipCellPainting |		// avoid drawing outside cell
					Tbl_smoothScrolling);		// easier to see the scrolling
					
	contents = new QList<QPixmap>;
	contents->setAutoDelete(true);
	
	setFrameStyle (QFrame::NoFrame);
	if (numCols < 1)	
		numCols = 1;
	setFixedWidth( (g_previewTableCellSize+6)*numCols+16);
	
	QObject::connect(&fillTimer, SIGNAL(timeout()), SLOT(fillTableAction()) );// allows table read/fill process to be interrupted.
}

ShaderPreview::~ShaderPreview()
{
	delete contents;
}


//------------------------------------------
void	ShaderPreview::setEnabled(bool status)
{
	enabled = status;
	if (enabled)
	{
		QTableView::setEnabled(true);
	}
	else
	{
		QTableView::setEnabled(false);
	}
}

//------------------------------------------
void ShaderPreview::changeNumCols(int newCols)
{
	maxCols = cols = newCols;
	setNumCols(cols);
	rows = (contents->count()+1)/cols;
	setNumRows(rows);
	update();
}

/**
 * Return content of cell
 */
QPixmap& ShaderPreview::cellContent( int row, int col )
{
    return *(contents->at( indexOf( row, col )) );	// contents array lookup
}


/**
 *  Delete table contents
 */
void ShaderPreview::clear()
{
	setNumRows(0);
	setNumCols(maxCols);
	rows = 0;
	cols = maxCols;
		
	curRow = curCol = lastRow = lastCol = -1;
	contents->clear();
}

/**
 *  set current position
 */
bool ShaderPreview::setCurrentItem (int row, int col)
{
	int	oldLastRow = lastRow;
	int	oldLastCol = lastCol;
	int oldCurRow = curRow;
	int oldCurCol = curCol;
	
	if (row < rows && col < cols)
	{
		curRow = lastRow = row;
		curCol = lastCol = col;
		updateCell( oldLastRow, oldLastCol );	// erase previous marking
		updateCell( oldCurRow, oldCurCol );		// erase previous marking
		updateCell( curRow, curCol );			// show new current cell
//		setTopCell(row);
		emit selected(indexOf(curRow, curCol));
		return true;
	}
	else
		return false;
}
bool ShaderPreview::setCurrentItem (int flat)
{
	return setCurrentItem( (flat/cols), (flat%cols) );
}

		
/**
 *  set content of cell
 */
void ShaderPreview::setCellContent( int row, int col, const QPixmap& pix )
{
		
	if (row+1 > rows)
	{
		setNumRows(row+1);
		rows = row+1;
	}
	if (col+1 > cols)
	{
		setNumCols(col+1);
		cols = col+1;
	}
		
	// create new pixmaps, if necessary
	for(int i = (int)contents->count(); i <= indexOf(row, col); i++)
		contents->append(new QPixmap);
	
    *(contents->at( indexOf(row, col) )) = pix;	// contents lookup and assign
	contents->at(indexOf( row, col ))->detach();
	updateCell( row, col );						// show new content
}

//------------------------------------------
void ShaderPreview::setCellContentNoUpdate( int row, int col, const QPixmap& pix )
{
	if (row+1 > rows)
	{
		setNumRows(row+1);
		rows = row+1;
	}
	if (col+1 > cols)
	{
		setNumCols(col+1);
		cols = col+1;
	}
	
	// create new pixmaps, if necessary
	for(int i = (int)contents->count(); i <= indexOf(row, col); i++)
		contents->append(new QPixmap);
	
    *(contents->at( indexOf(row, col) )) = pix;	// contents lookup and assign
	contents->at(indexOf( row, col ))->detach();
}

/**
 * Resize table bounds
 */
void ShaderPreview::setNumRowCols( int row, int col)
{
		
	if (row+1 > rows)
	{
		setNumRows(row+1);
		rows = row+1;
	}
	if (col+1 > cols)
	{
		setNumCols(col+1);
		cols = col+1;
	}
	
	// create new pixmaps, if necessary
	for(int i = (int)contents->count(); i <= indexOf(row, col); i++)
		contents->append(new QPixmap);
}

//------------------------------------------
void ShaderPreview::replace( int index, const QPixmap& pix, const char* cacheKey )
{
	int row;
	int col;
	int end;

	//FIXME -  THIS ISN'T WORKING?	
	// Insert or replace pixmap in pixmap cache...
	if(cacheKey)
		QPixmapCache::insert( cacheKey, pix );
		
	if (index < 0)
	{
		end = contents->count();
		row = end/cols;
		col = end%cols;
		setCellContent( row, col, pix );
		setSelected( end );
		setTopCell(row);
	}
	else if (index < (int)contents->count())
	{
		row = index/cols;
		col = index%cols;
		setCellContent( row, col, pix);
		setSelected( index );
	}
	else
	{
		warning("ShaderPreview::insert - trying to insert beyond count.");
		warning("count = %d, trying to insert at %d", (int)contents->count(), index);
	}
}

/**
 *  Handles cell painting for the Table widget.
 */
void ShaderPreview::paintCell( QPainter* p, int row, int col )
{
	if (row >= 0 && col >= 0)
	{
		int 	x2 = cellWidth( col ) - 2;
		int 	y2 = cellHeight( row ) - 2;
		QPixmap	*pix;
		int		index;

		index = indexOf(row,col);
		if (index < (int)contents->count() )
		{
			if(enabled)
			{
				//  Draw cell content
				pix = contents->at(index);
				if (pix)
					p->drawPixmap( 3, 3, *pix, 0, 0, 94, 94 );

				QPen	shadow(palette().normal().dark(), 2);
				QPen	hilight(palette().normal().light(), 1);
				p->setPen(hilight);
				p->drawRect( 2, 2, x2-2, y2-2 );
				p->setPen(shadow);
				p->moveTo (y2-1, 2);
				p->lineTo (2, 2);
				p->lineTo (2, x2-1);

				//-- highlight color is red (bright if in focus)
				if (hasFocus())
					p->setPen( QPen(QColor(255, 63, 63), 1));
				else
					p->setPen( QPen(QColor(159, 0, 31), 1) );

				// Draw extra frame inside if this is the current cell.
				if ( (row == lastRow) && (col == lastCol) )	// if we are on current cell,
				{
					p->drawRect( 0, 0, x2+2, y2+2 );			// draw rect. along cell edges
					p->drawRect( 2, 2, x2-2, y2-2 );			// draw rect. along cell edges
				}
				if ( (row == curRow) && (col == curCol) )	// if we are on current cell,
				{
					p->drawRect( 1, 1, x2, y2 );			// draw rect. along cell edges
				}
			}
			else
			{
				QBrush	bkGround(palette().normal().background());
				p->fillRect(0, 0, x2+2, y2+2, bkGround);
			}
		}
	}
}


/**
 * Handles mouse press events for the Table widget.
 * The current cell marker is set to the cell the mouse is clicked in.
 */
void ShaderPreview::mousePressEvent( QMouseEvent* e )
{
    int		oldRow = curRow;						// store previous current cell
    int		oldCol = curCol;

    QPoint clickedPos = e->pos();					// extract pointer position
	curRow = findRow(clickedPos.y());				// map to row; set current cell
	curCol = findCol(clickedPos.x());				// map to col; set current cell
	
	if ( curRow >= 0 && curCol >= 0 && indexOf(curRow, curCol) < (int)contents->count() ) // cell is valid 
	{
		if ( curRow != oldRow || curCol != oldCol )	// new cell selected
		{
			updateCell( oldRow, oldCol );			// erase previous marking
			updateCell( curRow, curCol );			// show new current cell
			emit highlighted(indexOf(curRow, curCol));
		}
		else
		{
			oldRow	= lastRow;
			oldCol	= lastCol;
			lastRow = curRow;
			lastCol = curCol;
			updateCell( oldRow, oldCol );			// erase previous marking
			updateCell( curRow, curCol );			// show new current cell
			emit selected(indexOf(curRow, curCol));
		}
	}
	else
	{
		curRow = oldRow;
		curCol = oldCol;
	}
}
//------------------------------------------
void ShaderPreview::setSelected( int index )
{
	int	oldCurRow	= curRow;
	int	oldCurCol	= curCol;
	int	oldLastRow	= lastRow;
	int	oldLastCol	= lastCol;

	if (index >= 0 && index < (int)contents->count() )
	{
		lastRow = curRow = index/cols;
		lastCol = curCol = index%cols;
		if ( curRow != oldCurRow || curCol != oldCurCol || lastRow != oldLastRow || lastCol != oldLastCol)
		{
			updateCell( oldCurRow, oldCurCol );
			updateCell( oldLastRow, oldLastCol );
			updateCell( curRow, curCol );
			updateCell( lastRow, lastCol );
			emit selected(index);
		}
	}
}

/**
 * Handles key press events for the Table widget.
 * Allows moving the current cell marker around with the arrow keys
 */
void ShaderPreview::keyPressEvent( QKeyEvent* e )
{
    int oldRow = curRow;				// store previous current cell
    int oldCol = curCol;
    int edge = 0;
    switch( e->key() ) {				// Look at the key code
	case Key_Left:						// If 'left arrow'-key, 
	    if( curCol > 0 ) {				// and cr't not in leftmost col
		curCol--;     					// set cr't to next left column
		edge = leftCell();				// find left edge
		if ( curCol < edge )			// if we have moved off  edge,
		    setLeftCell( edge - 1 );	// scroll view to rectify
	    }
	    break;
	case Key_Right:				// Correspondingly...
	    if( curCol < numCols()-1 ) {
		curCol++;
		edge = lastColVisible();
		if ( curCol >= edge )
		    setLeftCell( leftCell() + 1 );
	    }
	    break;
	case Key_Up:
	    if( curRow > 0 ) {
		curRow--;
		edge = topCell();
		if ( curRow < edge )
		    setTopCell( edge - 1 );
	    }
	    break;
	case Key_Down:
	    if( curRow < numRows()-1 ) {
		curRow++;
		edge = lastRowVisible();
		if ( curRow >= edge )
		    setTopCell( topCell() + 1 );
	    }
	    break;
	case Key_Space:
	case Key_Enter:	//something must be catching this because it's not getting to this widget.
		oldRow	= lastRow;
		oldCol	= lastCol;
		lastRow = curRow;
		lastCol = curCol;
		updateCell( oldRow, oldCol );			// erase previous marking
		updateCell( curRow, curCol );			// show new current cell
		emit selected(indexOf(curRow, curCol));
		oldRow	= curRow;		// set these so that it won't hit 
		oldCol	= curCol;		// the if statement below.
	    break;
	default:					// If not an interesting key,
	    e->ignore();			// we don't accept the event
	    return;	
    }
    
    if ( indexOf(curRow, curCol) < (int)contents->count()) 
    	if ( curRow!=oldRow || curCol!=oldCol ) // if current cell has moved,
		{
			updateCell( oldRow, oldCol );		// erase previous marking
			updateCell( curRow, curCol );		// show new current cell
			emit highlighted(indexOf(curRow, curCol));
    	}
	else
	{
		curRow = oldRow;
		curCol = oldCol;
	}
}

/**
 * Handles focus reception events for the Table widget.
 * Repaint only the current cell; to avoid flickering
 */
void ShaderPreview::focusInEvent( QFocusEvent* )
{
    updateCell( curRow, curCol );		// draw current cell
    updateCell( lastRow, lastCol );		// draw current cell
}    


/**
 * Handles focus loss events for the Table widget.
 * Repaint only the current cell; to avoid flickering
 */
void ShaderPreview::focusOutEvent( QFocusEvent* )
{
    updateCell( curRow, curCol );		// draw current cell
    updateCell( lastRow, lastCol );		// draw current cell
}    


/**
 * Utility function for mapping from 2D table to 1D array
 */
int ShaderPreview::indexOf( int row, int col )
{
    return (row * numCols()) + col;
}

//------------------------------------------
void ShaderPreview::fillTable( QCString& base, QStrList& previewFiles )
{
	if (previewFiles.count() < 1)
	{
		clear();
		update();
		return;
	}
	
	fillTimer.stop();
	
	prevFiles = previewFiles;
	baseDir = (const char*)base;
	
	fillCount = 0;
	fillMax = prevFiles.count();
	
	clear();
		
	setNumRowCols((fillMax-1)/cols, (fillMax-1)%cols);	//pre-create pixmaps to reduce flicker
	fillTimer.start(0);
}

//------------------------------------------
void ShaderPreview::fillTableAction( )
{
	QPixmap	*pTmpPix;
	
	if ( (pTmpPix = QPixmapCache::find(prevFiles.at(fillCount))) )
	{
		while (pTmpPix)
		{
			setCellContentNoUpdate( (fillCount)/cols, (fillCount)%cols, *pTmpPix);
			
			//--
			fillCount++;
			if (fillCount >= fillMax)
			{
				fillTimer.stop();
				pTmpPix = 0;
			}
			else 
				pTmpPix = QPixmapCache::find(prevFiles.at(fillCount));
		}
		update();
	}
	else
	{
		prevFilename = baseDir+prevFiles.at(fillCount);
		prevFilename += ".jpg";

		QPixmap	tmpPix((const char*)prevFilename, "JPEG");
		if (tmpPix.isNull())	// file didn't exist...use default image
			tmpPix = QPixmap( (const char**)noPreview_xpm );

		setCellContent( (fillCount)/cols, (fillCount)%cols, tmpPix);
		QPixmapCache::insert( prevFiles.at(fillCount), tmpPix );
		//--
		fillCount++;
		if (fillCount >= fillMax)
		{
			fillTimer.stop();
		}
	}
}
